# generated by datamodel-codegen:
#   filename:  schema.json
#   timestamp: 2025-11-11T09:15:18+00:00

from __future__ import annotations

from decimal import Decimal
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic.dataclasses import dataclass

from gcapi.model_base import BaseModel


@dataclass
class AlgorithmImage(BaseModel):
    pk: str
    url: str
    api_url: str
    algorithm: str
    created: str
    import_status: str
    image: str | None


@dataclass
class Answer(BaseModel):
    answer: Any | None
    api_url: str
    created: str
    creator: str
    display_set: str | None
    pk: str
    question: str
    modified: str
    answer_image: str | None
    last_edit_duration: str | None
    total_edit_duration: str | None
    is_ground_truth: bool


@dataclass
class AnswerRequest(BaseModel):
    answer: Any | None
    display_set: str | None
    question: str
    last_edit_duration: str | None


@dataclass
class Archive(BaseModel):
    pk: str
    title: str
    logo: str
    description: str | None
    api_url: str
    url: str


class BlankEnum(Enum):
    field_ = ""


@dataclass
class CategoricalOption(BaseModel):
    id: int
    title: str
    default: bool | None


@dataclass
class Challenge(BaseModel):
    title: str | None
    short_name: str


@dataclass
class ChallengeRequest(BaseModel):
    title: str | None
    short_name: str


class ColorInterpolationEnum(Enum):
    RGB = "RGB"
    HLS = "HLS"
    HLSpos = "HLSpos"
    HLSneg = "HLSneg"
    Constant = "Constant"


class ColorSpaceEnum(Enum):
    GRAY = "GRAY"
    RGB = "RGB"
    RGBA = "RGBA"
    YCBCR = "YCBCR"


@dataclass
class ComponentInterfaceValuePost(BaseModel):
    interface: str
    value: Any | None
    file: str | None
    image: str | None
    pk: int
    image_name: str | None


@dataclass
class ComponentInterfaceValuePostRequest(BaseModel):
    interface: str
    value: Any | None
    file: bytes | None
    image: str | None
    upload_session: str | None
    user_upload: str | None
    user_uploads: list[str] | None
    image_name: str | None


@dataclass
class DICOMImageSet(BaseModel):
    image_set_id: str


@dataclass
class DisplaySetPostRequest(BaseModel):
    reader_study: str | None
    values: list[ComponentInterfaceValuePostRequest] | None
    order: int | None
    title: str = ""


class EyeChoiceEnum(Enum):
    OD = "OD"
    OS = "OS"
    U = "U"
    NA = "NA"


@dataclass
class Feedback(BaseModel):
    session: str
    screenshot: str | None
    user_comment: str
    context: Any | None


@dataclass
class FeedbackRequest(BaseModel):
    session: str
    screenshot: bytes | None
    user_comment: str
    context: Any | None


class FieldOfViewEnum(Enum):
    F1M = "F1M"
    F2 = "F2"
    F3M = "F3M"
    F4 = "F4"
    F5 = "F5"
    F6 = "F6"
    F7 = "F7"
    U = "U"


@dataclass
class Follow(BaseModel):
    pk: int
    flag: str | None


@dataclass
class FollowRequest(BaseModel):
    flag: str | None


@dataclass
class GCAPIVersion(BaseModel):
    latest_version: str
    lowest_supported_version: str


@dataclass
class HangingProtocol(BaseModel):
    json: Any
    title: str
    pk: str
    svg_icon: str


@dataclass
class HangingProtocolRequest(BaseModel):
    json: Any
    title: str


class ImageTypeEnum(Enum):
    MHD = "MHD"
    TIFF = "TIFF"
    DZI = "DZI"


@dataclass
class ImagingModality(BaseModel):
    modality: str


@dataclass
class JobPost(BaseModel):
    pk: str
    inputs: list[ComponentInterfaceValuePost]
    status: str


@dataclass
class JobPostRequest(BaseModel):
    algorithm: str
    inputs: list[ComponentInterfaceValuePostRequest]


class LocationEnum(Enum):
    AF = "AF"
    AX = "AX"
    AL = "AL"
    DZ = "DZ"
    AS = "AS"
    AD = "AD"
    AO = "AO"
    AI = "AI"
    AQ = "AQ"
    AG = "AG"
    AR = "AR"
    AM = "AM"
    AW = "AW"
    AU = "AU"
    AT = "AT"
    AZ = "AZ"
    BS = "BS"
    BH = "BH"
    BD = "BD"
    BB = "BB"
    BY = "BY"
    BE = "BE"
    BZ = "BZ"
    BJ = "BJ"
    BM = "BM"
    BT = "BT"
    BO = "BO"
    BQ = "BQ"
    BA = "BA"
    BW = "BW"
    BV = "BV"
    BR = "BR"
    IO = "IO"
    BN = "BN"
    BG = "BG"
    BF = "BF"
    BI = "BI"
    CV = "CV"
    KH = "KH"
    CM = "CM"
    CA = "CA"
    KY = "KY"
    CF = "CF"
    TD = "TD"
    CL = "CL"
    CN = "CN"
    CX = "CX"
    CC = "CC"
    CO = "CO"
    KM = "KM"
    CG = "CG"
    CK = "CK"
    CR = "CR"
    CI = "CI"
    HR = "HR"
    CU = "CU"
    CW = "CW"
    CY = "CY"
    CZ = "CZ"
    CD = "CD"
    DK = "DK"
    DJ = "DJ"
    DM = "DM"
    DO = "DO"
    EC = "EC"
    EG = "EG"
    SV = "SV"
    GQ = "GQ"
    ER = "ER"
    EE = "EE"
    SZ = "SZ"
    ET = "ET"
    FK = "FK"
    FO = "FO"
    FJ = "FJ"
    FI = "FI"
    FR = "FR"
    GF = "GF"
    PF = "PF"
    TF = "TF"
    GA = "GA"
    GM = "GM"
    GE = "GE"
    DE = "DE"
    GH = "GH"
    GI = "GI"
    GR = "GR"
    GL = "GL"
    GD = "GD"
    GP = "GP"
    GU = "GU"
    GT = "GT"
    GG = "GG"
    GN = "GN"
    GW = "GW"
    GY = "GY"
    HT = "HT"
    HM = "HM"
    HN = "HN"
    HK = "HK"
    HU = "HU"
    IS = "IS"
    IN = "IN"
    ID = "ID"
    IR = "IR"
    IQ = "IQ"
    IE = "IE"
    IM = "IM"
    IL = "IL"
    IT = "IT"
    JM = "JM"
    JP = "JP"
    JE = "JE"
    JO = "JO"
    KZ = "KZ"
    KE = "KE"
    KI = "KI"
    KW = "KW"
    KG = "KG"
    LA = "LA"
    LV = "LV"
    LB = "LB"
    LS = "LS"
    LR = "LR"
    LY = "LY"
    LI = "LI"
    LT = "LT"
    LU = "LU"
    MO = "MO"
    MG = "MG"
    MW = "MW"
    MY = "MY"
    MV = "MV"
    ML = "ML"
    MT = "MT"
    MH = "MH"
    MQ = "MQ"
    MR = "MR"
    MU = "MU"
    YT = "YT"
    MX = "MX"
    FM = "FM"
    MD = "MD"
    MC = "MC"
    MN = "MN"
    ME = "ME"
    MS = "MS"
    MA = "MA"
    MZ = "MZ"
    MM = "MM"
    NA = "NA"
    NR = "NR"
    NP = "NP"
    NL = "NL"
    NC = "NC"
    NZ = "NZ"
    NI = "NI"
    NE = "NE"
    NG = "NG"
    NU = "NU"
    NF = "NF"
    KP = "KP"
    MK = "MK"
    MP = "MP"
    NO = "NO"
    OM = "OM"
    PK = "PK"
    PW = "PW"
    PS = "PS"
    PA = "PA"
    PG = "PG"
    PY = "PY"
    PE = "PE"
    PH = "PH"
    PN = "PN"
    PL = "PL"
    PT = "PT"
    PR = "PR"
    QA = "QA"
    RE = "RE"
    RO = "RO"
    RU = "RU"
    RW = "RW"
    BL = "BL"
    SH = "SH"
    KN = "KN"
    LC = "LC"
    MF = "MF"
    PM = "PM"
    VC = "VC"
    WS = "WS"
    SM = "SM"
    ST = "ST"
    SA = "SA"
    SN = "SN"
    RS = "RS"
    SC = "SC"
    SL = "SL"
    SG = "SG"
    SX = "SX"
    SK = "SK"
    SI = "SI"
    SB = "SB"
    SO = "SO"
    ZA = "ZA"
    GS = "GS"
    KR = "KR"
    SS = "SS"
    ES = "ES"
    LK = "LK"
    SD = "SD"
    SR = "SR"
    SJ = "SJ"
    SE = "SE"
    CH = "CH"
    SY = "SY"
    TW = "TW"
    TJ = "TJ"
    TZ = "TZ"
    TH = "TH"
    TL = "TL"
    TG = "TG"
    TK = "TK"
    TO = "TO"
    TT = "TT"
    TN = "TN"
    TR = "TR"
    TM = "TM"
    TC = "TC"
    TV = "TV"
    UG = "UG"
    UA = "UA"
    AE = "AE"
    GB = "GB"
    UM = "UM"
    US = "US"
    UY = "UY"
    UZ = "UZ"
    VU = "VU"
    VA = "VA"
    VE = "VE"
    VN = "VN"
    VG = "VG"
    VI = "VI"
    WF = "WF"
    EH = "EH"
    YE = "YE"
    ZM = "ZM"
    ZW = "ZW"


@dataclass
class LookUpTable(BaseModel):
    pk: int
    slug: str
    title: str
    description: str | None
    color: str
    alpha: str
    color_invert: str | None
    alpha_invert: str | None
    range_min: int | None
    range_max: int | None
    relative: bool | None
    color_interpolation: ColorInterpolationEnum | None
    color_interpolation_invert: ColorInterpolationEnum | None


@dataclass
class Notification(BaseModel):
    read: bool | None


@dataclass
class NotificationRequest(BaseModel):
    read: bool | None


class NullEnum(Enum):
    NoneType_None = None


@dataclass
class PaginatedAlgorithmImageList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[AlgorithmImage]


@dataclass
class PaginatedAnswerList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Answer]


@dataclass
class PaginatedArchiveList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Archive]


@dataclass
class PaginatedFeedbackList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Feedback]


@dataclass
class PaginatedFollowList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Follow]


@dataclass
class PaginatedNotificationList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Notification]


@dataclass
class Part(BaseModel):
    ETag: str
    PartNumber: int
    LastModified: str
    Size: int


@dataclass
class PartRequest(BaseModel):
    ETag: str
    PartNumber: int
    LastModified: str
    Size: int


@dataclass
class PatchedAnswerRequest(BaseModel):
    answer: Any | None
    display_set: str | None
    question: str | None
    last_edit_duration: str | None


@dataclass
class PatchedArchiveItemPostRequest(BaseModel):
    archive: str | None
    values: list[ComponentInterfaceValuePostRequest] | None
    title: str = ""


@dataclass
class PatchedDisplaySetPostRequest(BaseModel):
    reader_study: str | None
    values: list[ComponentInterfaceValuePostRequest] | None
    order: int | None
    title: str = ""


@dataclass
class PatchedFollowRequest(BaseModel):
    flag: str | None


@dataclass
class PatchedNotificationRequest(BaseModel):
    read: bool | None


@dataclass
class PatchedUserUploadCompleteRequest(BaseModel):
    parts: list[PartRequest] | None


@dataclass
class PatchedUserUploadPresignedURLsRequest(BaseModel):
    part_numbers: list[int] | None


class PatientSexEnum(Enum):
    M = "M"
    F = "F"
    O = "O"


@dataclass
class Phase(BaseModel):
    challenge: Challenge
    title: str
    slug: str


@dataclass
class PhaseRequest(BaseModel):
    challenge: ChallengeRequest
    title: str


@dataclass
class PublicChallenge(BaseModel):
    api_url: str
    url: str
    slug: str
    title: str | None
    description: str | None
    public: bool
    status: str
    logo: str | None
    submission_types: list[str]
    start_date: str | None
    end_date: str | None
    publications: list[str]
    created: str
    modified: str
    incentives: list[str]


@dataclass
class RawImageUploadSession(BaseModel):
    pk: str
    creator: int | None
    status: str
    error_message: str | None
    image_set: list[str]
    api_url: str
    user_uploads: list[str] | None
    uploads: list[str]


@dataclass
class RawImageUploadSessionRequest(BaseModel):
    creator: int | None
    error_message: str | None
    user_uploads: list[str] | None
    archive: str | None
    answer: str | None
    interface: str | None
    archive_item: str | None
    display_set: str | None
    uploads: list[str]


@dataclass
class Session(BaseModel):
    pk: str
    status: str


@dataclass
class SimpleImage(BaseModel):
    pk: str
    name: str


class StatusEnum(Enum):
    Succeeded = "Succeeded"
    Failed = "Failed"


class StereoscopicChoiceEnum(Enum):
    L = "L"
    R = "R"
    U = "U"


class TimezoneEnum(Enum):
    Africa_Abidjan = "Africa/Abidjan"
    Africa_Accra = "Africa/Accra"
    Africa_Addis_Ababa = "Africa/Addis_Ababa"
    Africa_Algiers = "Africa/Algiers"
    Africa_Asmara = "Africa/Asmara"
    Africa_Asmera = "Africa/Asmera"
    Africa_Bamako = "Africa/Bamako"
    Africa_Bangui = "Africa/Bangui"
    Africa_Banjul = "Africa/Banjul"
    Africa_Bissau = "Africa/Bissau"
    Africa_Blantyre = "Africa/Blantyre"
    Africa_Brazzaville = "Africa/Brazzaville"
    Africa_Bujumbura = "Africa/Bujumbura"
    Africa_Cairo = "Africa/Cairo"
    Africa_Casablanca = "Africa/Casablanca"
    Africa_Ceuta = "Africa/Ceuta"
    Africa_Conakry = "Africa/Conakry"
    Africa_Dakar = "Africa/Dakar"
    Africa_Dar_es_Salaam = "Africa/Dar_es_Salaam"
    Africa_Djibouti = "Africa/Djibouti"
    Africa_Douala = "Africa/Douala"
    Africa_El_Aaiun = "Africa/El_Aaiun"
    Africa_Freetown = "Africa/Freetown"
    Africa_Gaborone = "Africa/Gaborone"
    Africa_Harare = "Africa/Harare"
    Africa_Johannesburg = "Africa/Johannesburg"
    Africa_Juba = "Africa/Juba"
    Africa_Kampala = "Africa/Kampala"
    Africa_Khartoum = "Africa/Khartoum"
    Africa_Kigali = "Africa/Kigali"
    Africa_Kinshasa = "Africa/Kinshasa"
    Africa_Lagos = "Africa/Lagos"
    Africa_Libreville = "Africa/Libreville"
    Africa_Lome = "Africa/Lome"
    Africa_Luanda = "Africa/Luanda"
    Africa_Lubumbashi = "Africa/Lubumbashi"
    Africa_Lusaka = "Africa/Lusaka"
    Africa_Malabo = "Africa/Malabo"
    Africa_Maputo = "Africa/Maputo"
    Africa_Maseru = "Africa/Maseru"
    Africa_Mbabane = "Africa/Mbabane"
    Africa_Mogadishu = "Africa/Mogadishu"
    Africa_Monrovia = "Africa/Monrovia"
    Africa_Nairobi = "Africa/Nairobi"
    Africa_Ndjamena = "Africa/Ndjamena"
    Africa_Niamey = "Africa/Niamey"
    Africa_Nouakchott = "Africa/Nouakchott"
    Africa_Ouagadougou = "Africa/Ouagadougou"
    Africa_Porto_Novo = "Africa/Porto-Novo"
    Africa_Sao_Tome = "Africa/Sao_Tome"
    Africa_Timbuktu = "Africa/Timbuktu"
    Africa_Tripoli = "Africa/Tripoli"
    Africa_Tunis = "Africa/Tunis"
    Africa_Windhoek = "Africa/Windhoek"
    America_Adak = "America/Adak"
    America_Anchorage = "America/Anchorage"
    America_Anguilla = "America/Anguilla"
    America_Antigua = "America/Antigua"
    America_Araguaina = "America/Araguaina"
    America_Argentina_Buenos_Aires = "America/Argentina/Buenos_Aires"
    America_Argentina_Catamarca = "America/Argentina/Catamarca"
    America_Argentina_ComodRivadavia = "America/Argentina/ComodRivadavia"
    America_Argentina_Cordoba = "America/Argentina/Cordoba"
    America_Argentina_Jujuy = "America/Argentina/Jujuy"
    America_Argentina_La_Rioja = "America/Argentina/La_Rioja"
    America_Argentina_Mendoza = "America/Argentina/Mendoza"
    America_Argentina_Rio_Gallegos = "America/Argentina/Rio_Gallegos"
    America_Argentina_Salta = "America/Argentina/Salta"
    America_Argentina_San_Juan = "America/Argentina/San_Juan"
    America_Argentina_San_Luis = "America/Argentina/San_Luis"
    America_Argentina_Tucuman = "America/Argentina/Tucuman"
    America_Argentina_Ushuaia = "America/Argentina/Ushuaia"
    America_Aruba = "America/Aruba"
    America_Asuncion = "America/Asuncion"
    America_Atikokan = "America/Atikokan"
    America_Atka = "America/Atka"
    America_Bahia = "America/Bahia"
    America_Bahia_Banderas = "America/Bahia_Banderas"
    America_Barbados = "America/Barbados"
    America_Belem = "America/Belem"
    America_Belize = "America/Belize"
    America_Blanc_Sablon = "America/Blanc-Sablon"
    America_Boa_Vista = "America/Boa_Vista"
    America_Bogota = "America/Bogota"
    America_Boise = "America/Boise"
    America_Buenos_Aires = "America/Buenos_Aires"
    America_Cambridge_Bay = "America/Cambridge_Bay"
    America_Campo_Grande = "America/Campo_Grande"
    America_Cancun = "America/Cancun"
    America_Caracas = "America/Caracas"
    America_Catamarca = "America/Catamarca"
    America_Cayenne = "America/Cayenne"
    America_Cayman = "America/Cayman"
    America_Chicago = "America/Chicago"
    America_Chihuahua = "America/Chihuahua"
    America_Ciudad_Juarez = "America/Ciudad_Juarez"
    America_Coral_Harbour = "America/Coral_Harbour"
    America_Cordoba = "America/Cordoba"
    America_Costa_Rica = "America/Costa_Rica"
    America_Coyhaique = "America/Coyhaique"
    America_Creston = "America/Creston"
    America_Cuiaba = "America/Cuiaba"
    America_Curacao = "America/Curacao"
    America_Danmarkshavn = "America/Danmarkshavn"
    America_Dawson = "America/Dawson"
    America_Dawson_Creek = "America/Dawson_Creek"
    America_Denver = "America/Denver"
    America_Detroit = "America/Detroit"
    America_Dominica = "America/Dominica"
    America_Edmonton = "America/Edmonton"
    America_Eirunepe = "America/Eirunepe"
    America_El_Salvador = "America/El_Salvador"
    America_Ensenada = "America/Ensenada"
    America_Fort_Nelson = "America/Fort_Nelson"
    America_Fort_Wayne = "America/Fort_Wayne"
    America_Fortaleza = "America/Fortaleza"
    America_Glace_Bay = "America/Glace_Bay"
    America_Godthab = "America/Godthab"
    America_Goose_Bay = "America/Goose_Bay"
    America_Grand_Turk = "America/Grand_Turk"
    America_Grenada = "America/Grenada"
    America_Guadeloupe = "America/Guadeloupe"
    America_Guatemala = "America/Guatemala"
    America_Guayaquil = "America/Guayaquil"
    America_Guyana = "America/Guyana"
    America_Halifax = "America/Halifax"
    America_Havana = "America/Havana"
    America_Hermosillo = "America/Hermosillo"
    America_Indiana_Indianapolis = "America/Indiana/Indianapolis"
    America_Indiana_Knox = "America/Indiana/Knox"
    America_Indiana_Marengo = "America/Indiana/Marengo"
    America_Indiana_Petersburg = "America/Indiana/Petersburg"
    America_Indiana_Tell_City = "America/Indiana/Tell_City"
    America_Indiana_Vevay = "America/Indiana/Vevay"
    America_Indiana_Vincennes = "America/Indiana/Vincennes"
    America_Indiana_Winamac = "America/Indiana/Winamac"
    America_Indianapolis = "America/Indianapolis"
    America_Inuvik = "America/Inuvik"
    America_Iqaluit = "America/Iqaluit"
    America_Jamaica = "America/Jamaica"
    America_Jujuy = "America/Jujuy"
    America_Juneau = "America/Juneau"
    America_Kentucky_Louisville = "America/Kentucky/Louisville"
    America_Kentucky_Monticello = "America/Kentucky/Monticello"
    America_Knox_IN = "America/Knox_IN"
    America_Kralendijk = "America/Kralendijk"
    America_La_Paz = "America/La_Paz"
    America_Lima = "America/Lima"
    America_Los_Angeles = "America/Los_Angeles"
    America_Louisville = "America/Louisville"
    America_Lower_Princes = "America/Lower_Princes"
    America_Maceio = "America/Maceio"
    America_Managua = "America/Managua"
    America_Manaus = "America/Manaus"
    America_Marigot = "America/Marigot"
    America_Martinique = "America/Martinique"
    America_Matamoros = "America/Matamoros"
    America_Mazatlan = "America/Mazatlan"
    America_Mendoza = "America/Mendoza"
    America_Menominee = "America/Menominee"
    America_Merida = "America/Merida"
    America_Metlakatla = "America/Metlakatla"
    America_Mexico_City = "America/Mexico_City"
    America_Miquelon = "America/Miquelon"
    America_Moncton = "America/Moncton"
    America_Monterrey = "America/Monterrey"
    America_Montevideo = "America/Montevideo"
    America_Montreal = "America/Montreal"
    America_Montserrat = "America/Montserrat"
    America_Nassau = "America/Nassau"
    America_New_York = "America/New_York"
    America_Nipigon = "America/Nipigon"
    America_Nome = "America/Nome"
    America_Noronha = "America/Noronha"
    America_North_Dakota_Beulah = "America/North_Dakota/Beulah"
    America_North_Dakota_Center = "America/North_Dakota/Center"
    America_North_Dakota_New_Salem = "America/North_Dakota/New_Salem"
    America_Nuuk = "America/Nuuk"
    America_Ojinaga = "America/Ojinaga"
    America_Panama = "America/Panama"
    America_Pangnirtung = "America/Pangnirtung"
    America_Paramaribo = "America/Paramaribo"
    America_Phoenix = "America/Phoenix"
    America_Port_au_Prince = "America/Port-au-Prince"
    America_Port_of_Spain = "America/Port_of_Spain"
    America_Porto_Acre = "America/Porto_Acre"
    America_Porto_Velho = "America/Porto_Velho"
    America_Puerto_Rico = "America/Puerto_Rico"
    America_Punta_Arenas = "America/Punta_Arenas"
    America_Rainy_River = "America/Rainy_River"
    America_Rankin_Inlet = "America/Rankin_Inlet"
    America_Recife = "America/Recife"
    America_Regina = "America/Regina"
    America_Resolute = "America/Resolute"
    America_Rio_Branco = "America/Rio_Branco"
    America_Rosario = "America/Rosario"
    America_Santa_Isabel = "America/Santa_Isabel"
    America_Santarem = "America/Santarem"
    America_Santiago = "America/Santiago"
    America_Santo_Domingo = "America/Santo_Domingo"
    America_Sao_Paulo = "America/Sao_Paulo"
    America_Scoresbysund = "America/Scoresbysund"
    America_Shiprock = "America/Shiprock"
    America_Sitka = "America/Sitka"
    America_St_Barthelemy = "America/St_Barthelemy"
    America_St_Johns = "America/St_Johns"
    America_St_Kitts = "America/St_Kitts"
    America_St_Lucia = "America/St_Lucia"
    America_St_Thomas = "America/St_Thomas"
    America_St_Vincent = "America/St_Vincent"
    America_Swift_Current = "America/Swift_Current"
    America_Tegucigalpa = "America/Tegucigalpa"
    America_Thule = "America/Thule"
    America_Thunder_Bay = "America/Thunder_Bay"
    America_Tijuana = "America/Tijuana"
    America_Toronto = "America/Toronto"
    America_Tortola = "America/Tortola"
    America_Vancouver = "America/Vancouver"
    America_Virgin = "America/Virgin"
    America_Whitehorse = "America/Whitehorse"
    America_Winnipeg = "America/Winnipeg"
    America_Yakutat = "America/Yakutat"
    America_Yellowknife = "America/Yellowknife"
    Antarctica_Casey = "Antarctica/Casey"
    Antarctica_Davis = "Antarctica/Davis"
    Antarctica_DumontDUrville = "Antarctica/DumontDUrville"
    Antarctica_Macquarie = "Antarctica/Macquarie"
    Antarctica_Mawson = "Antarctica/Mawson"
    Antarctica_McMurdo = "Antarctica/McMurdo"
    Antarctica_Palmer = "Antarctica/Palmer"
    Antarctica_Rothera = "Antarctica/Rothera"
    Antarctica_South_Pole = "Antarctica/South_Pole"
    Antarctica_Syowa = "Antarctica/Syowa"
    Antarctica_Troll = "Antarctica/Troll"
    Antarctica_Vostok = "Antarctica/Vostok"
    Arctic_Longyearbyen = "Arctic/Longyearbyen"
    Asia_Aden = "Asia/Aden"
    Asia_Almaty = "Asia/Almaty"
    Asia_Amman = "Asia/Amman"
    Asia_Anadyr = "Asia/Anadyr"
    Asia_Aqtau = "Asia/Aqtau"
    Asia_Aqtobe = "Asia/Aqtobe"
    Asia_Ashgabat = "Asia/Ashgabat"
    Asia_Ashkhabad = "Asia/Ashkhabad"
    Asia_Atyrau = "Asia/Atyrau"
    Asia_Baghdad = "Asia/Baghdad"
    Asia_Bahrain = "Asia/Bahrain"
    Asia_Baku = "Asia/Baku"
    Asia_Bangkok = "Asia/Bangkok"
    Asia_Barnaul = "Asia/Barnaul"
    Asia_Beirut = "Asia/Beirut"
    Asia_Bishkek = "Asia/Bishkek"
    Asia_Brunei = "Asia/Brunei"
    Asia_Calcutta = "Asia/Calcutta"
    Asia_Chita = "Asia/Chita"
    Asia_Choibalsan = "Asia/Choibalsan"
    Asia_Chongqing = "Asia/Chongqing"
    Asia_Chungking = "Asia/Chungking"
    Asia_Colombo = "Asia/Colombo"
    Asia_Dacca = "Asia/Dacca"
    Asia_Damascus = "Asia/Damascus"
    Asia_Dhaka = "Asia/Dhaka"
    Asia_Dili = "Asia/Dili"
    Asia_Dubai = "Asia/Dubai"
    Asia_Dushanbe = "Asia/Dushanbe"
    Asia_Famagusta = "Asia/Famagusta"
    Asia_Gaza = "Asia/Gaza"
    Asia_Harbin = "Asia/Harbin"
    Asia_Hebron = "Asia/Hebron"
    Asia_Ho_Chi_Minh = "Asia/Ho_Chi_Minh"
    Asia_Hong_Kong = "Asia/Hong_Kong"
    Asia_Hovd = "Asia/Hovd"
    Asia_Irkutsk = "Asia/Irkutsk"
    Asia_Istanbul = "Asia/Istanbul"
    Asia_Jakarta = "Asia/Jakarta"
    Asia_Jayapura = "Asia/Jayapura"
    Asia_Jerusalem = "Asia/Jerusalem"
    Asia_Kabul = "Asia/Kabul"
    Asia_Kamchatka = "Asia/Kamchatka"
    Asia_Karachi = "Asia/Karachi"
    Asia_Kashgar = "Asia/Kashgar"
    Asia_Kathmandu = "Asia/Kathmandu"
    Asia_Katmandu = "Asia/Katmandu"
    Asia_Khandyga = "Asia/Khandyga"
    Asia_Kolkata = "Asia/Kolkata"
    Asia_Krasnoyarsk = "Asia/Krasnoyarsk"
    Asia_Kuala_Lumpur = "Asia/Kuala_Lumpur"
    Asia_Kuching = "Asia/Kuching"
    Asia_Kuwait = "Asia/Kuwait"
    Asia_Macao = "Asia/Macao"
    Asia_Macau = "Asia/Macau"
    Asia_Magadan = "Asia/Magadan"
    Asia_Makassar = "Asia/Makassar"
    Asia_Manila = "Asia/Manila"
    Asia_Muscat = "Asia/Muscat"
    Asia_Nicosia = "Asia/Nicosia"
    Asia_Novokuznetsk = "Asia/Novokuznetsk"
    Asia_Novosibirsk = "Asia/Novosibirsk"
    Asia_Omsk = "Asia/Omsk"
    Asia_Oral = "Asia/Oral"
    Asia_Phnom_Penh = "Asia/Phnom_Penh"
    Asia_Pontianak = "Asia/Pontianak"
    Asia_Pyongyang = "Asia/Pyongyang"
    Asia_Qatar = "Asia/Qatar"
    Asia_Qostanay = "Asia/Qostanay"
    Asia_Qyzylorda = "Asia/Qyzylorda"
    Asia_Rangoon = "Asia/Rangoon"
    Asia_Riyadh = "Asia/Riyadh"
    Asia_Saigon = "Asia/Saigon"
    Asia_Sakhalin = "Asia/Sakhalin"
    Asia_Samarkand = "Asia/Samarkand"
    Asia_Seoul = "Asia/Seoul"
    Asia_Shanghai = "Asia/Shanghai"
    Asia_Singapore = "Asia/Singapore"
    Asia_Srednekolymsk = "Asia/Srednekolymsk"
    Asia_Taipei = "Asia/Taipei"
    Asia_Tashkent = "Asia/Tashkent"
    Asia_Tbilisi = "Asia/Tbilisi"
    Asia_Tehran = "Asia/Tehran"
    Asia_Tel_Aviv = "Asia/Tel_Aviv"
    Asia_Thimbu = "Asia/Thimbu"
    Asia_Thimphu = "Asia/Thimphu"
    Asia_Tokyo = "Asia/Tokyo"
    Asia_Tomsk = "Asia/Tomsk"
    Asia_Ujung_Pandang = "Asia/Ujung_Pandang"
    Asia_Ulaanbaatar = "Asia/Ulaanbaatar"
    Asia_Ulan_Bator = "Asia/Ulan_Bator"
    Asia_Urumqi = "Asia/Urumqi"
    Asia_Ust_Nera = "Asia/Ust-Nera"
    Asia_Vientiane = "Asia/Vientiane"
    Asia_Vladivostok = "Asia/Vladivostok"
    Asia_Yakutsk = "Asia/Yakutsk"
    Asia_Yangon = "Asia/Yangon"
    Asia_Yekaterinburg = "Asia/Yekaterinburg"
    Asia_Yerevan = "Asia/Yerevan"
    Atlantic_Azores = "Atlantic/Azores"
    Atlantic_Bermuda = "Atlantic/Bermuda"
    Atlantic_Canary = "Atlantic/Canary"
    Atlantic_Cape_Verde = "Atlantic/Cape_Verde"
    Atlantic_Faeroe = "Atlantic/Faeroe"
    Atlantic_Faroe = "Atlantic/Faroe"
    Atlantic_Jan_Mayen = "Atlantic/Jan_Mayen"
    Atlantic_Madeira = "Atlantic/Madeira"
    Atlantic_Reykjavik = "Atlantic/Reykjavik"
    Atlantic_South_Georgia = "Atlantic/South_Georgia"
    Atlantic_St_Helena = "Atlantic/St_Helena"
    Atlantic_Stanley = "Atlantic/Stanley"
    Australia_ACT = "Australia/ACT"
    Australia_Adelaide = "Australia/Adelaide"
    Australia_Brisbane = "Australia/Brisbane"
    Australia_Broken_Hill = "Australia/Broken_Hill"
    Australia_Canberra = "Australia/Canberra"
    Australia_Currie = "Australia/Currie"
    Australia_Darwin = "Australia/Darwin"
    Australia_Eucla = "Australia/Eucla"
    Australia_Hobart = "Australia/Hobart"
    Australia_LHI = "Australia/LHI"
    Australia_Lindeman = "Australia/Lindeman"
    Australia_Lord_Howe = "Australia/Lord_Howe"
    Australia_Melbourne = "Australia/Melbourne"
    Australia_NSW = "Australia/NSW"
    Australia_North = "Australia/North"
    Australia_Perth = "Australia/Perth"
    Australia_Queensland = "Australia/Queensland"
    Australia_South = "Australia/South"
    Australia_Sydney = "Australia/Sydney"
    Australia_Tasmania = "Australia/Tasmania"
    Australia_Victoria = "Australia/Victoria"
    Australia_West = "Australia/West"
    Australia_Yancowinna = "Australia/Yancowinna"
    Brazil_Acre = "Brazil/Acre"
    Brazil_DeNoronha = "Brazil/DeNoronha"
    Brazil_East = "Brazil/East"
    Brazil_West = "Brazil/West"
    CET = "CET"
    CST6CDT = "CST6CDT"
    Canada_Atlantic = "Canada/Atlantic"
    Canada_Central = "Canada/Central"
    Canada_Eastern = "Canada/Eastern"
    Canada_Mountain = "Canada/Mountain"
    Canada_Newfoundland = "Canada/Newfoundland"
    Canada_Pacific = "Canada/Pacific"
    Canada_Saskatchewan = "Canada/Saskatchewan"
    Canada_Yukon = "Canada/Yukon"
    Chile_Continental = "Chile/Continental"
    Chile_EasterIsland = "Chile/EasterIsland"
    Cuba = "Cuba"
    EET = "EET"
    EST = "EST"
    EST5EDT = "EST5EDT"
    Egypt = "Egypt"
    Eire = "Eire"
    Etc_GMT = "Etc/GMT"
    Etc_GMT_0 = "Etc/GMT+0"
    Etc_GMT_1 = "Etc/GMT+1"
    Etc_GMT_10 = "Etc/GMT+10"
    Etc_GMT_11 = "Etc/GMT+11"
    Etc_GMT_12 = "Etc/GMT+12"
    Etc_GMT_2 = "Etc/GMT+2"
    Etc_GMT_3 = "Etc/GMT+3"
    Etc_GMT_4 = "Etc/GMT+4"
    Etc_GMT_5 = "Etc/GMT+5"
    Etc_GMT_6 = "Etc/GMT+6"
    Etc_GMT_7 = "Etc/GMT+7"
    Etc_GMT_8 = "Etc/GMT+8"
    Etc_GMT_9 = "Etc/GMT+9"
    Etc_GMT_0_1 = "Etc/GMT-0"
    Etc_GMT_1_1 = "Etc/GMT-1"
    Etc_GMT_10_1 = "Etc/GMT-10"
    Etc_GMT_11_1 = "Etc/GMT-11"
    Etc_GMT_12_1 = "Etc/GMT-12"
    Etc_GMT_13 = "Etc/GMT-13"
    Etc_GMT_14 = "Etc/GMT-14"
    Etc_GMT_2_1 = "Etc/GMT-2"
    Etc_GMT_3_1 = "Etc/GMT-3"
    Etc_GMT_4_1 = "Etc/GMT-4"
    Etc_GMT_5_1 = "Etc/GMT-5"
    Etc_GMT_6_1 = "Etc/GMT-6"
    Etc_GMT_7_1 = "Etc/GMT-7"
    Etc_GMT_8_1 = "Etc/GMT-8"
    Etc_GMT_9_1 = "Etc/GMT-9"
    Etc_GMT0 = "Etc/GMT0"
    Etc_Greenwich = "Etc/Greenwich"
    Etc_UCT = "Etc/UCT"
    Etc_UTC = "Etc/UTC"
    Etc_Universal = "Etc/Universal"
    Etc_Zulu = "Etc/Zulu"
    Europe_Amsterdam = "Europe/Amsterdam"
    Europe_Andorra = "Europe/Andorra"
    Europe_Astrakhan = "Europe/Astrakhan"
    Europe_Athens = "Europe/Athens"
    Europe_Belfast = "Europe/Belfast"
    Europe_Belgrade = "Europe/Belgrade"
    Europe_Berlin = "Europe/Berlin"
    Europe_Bratislava = "Europe/Bratislava"
    Europe_Brussels = "Europe/Brussels"
    Europe_Bucharest = "Europe/Bucharest"
    Europe_Budapest = "Europe/Budapest"
    Europe_Busingen = "Europe/Busingen"
    Europe_Chisinau = "Europe/Chisinau"
    Europe_Copenhagen = "Europe/Copenhagen"
    Europe_Dublin = "Europe/Dublin"
    Europe_Gibraltar = "Europe/Gibraltar"
    Europe_Guernsey = "Europe/Guernsey"
    Europe_Helsinki = "Europe/Helsinki"
    Europe_Isle_of_Man = "Europe/Isle_of_Man"
    Europe_Istanbul = "Europe/Istanbul"
    Europe_Jersey = "Europe/Jersey"
    Europe_Kaliningrad = "Europe/Kaliningrad"
    Europe_Kiev = "Europe/Kiev"
    Europe_Kirov = "Europe/Kirov"
    Europe_Kyiv = "Europe/Kyiv"
    Europe_Lisbon = "Europe/Lisbon"
    Europe_Ljubljana = "Europe/Ljubljana"
    Europe_London = "Europe/London"
    Europe_Luxembourg = "Europe/Luxembourg"
    Europe_Madrid = "Europe/Madrid"
    Europe_Malta = "Europe/Malta"
    Europe_Mariehamn = "Europe/Mariehamn"
    Europe_Minsk = "Europe/Minsk"
    Europe_Monaco = "Europe/Monaco"
    Europe_Moscow = "Europe/Moscow"
    Europe_Nicosia = "Europe/Nicosia"
    Europe_Oslo = "Europe/Oslo"
    Europe_Paris = "Europe/Paris"
    Europe_Podgorica = "Europe/Podgorica"
    Europe_Prague = "Europe/Prague"
    Europe_Riga = "Europe/Riga"
    Europe_Rome = "Europe/Rome"
    Europe_Samara = "Europe/Samara"
    Europe_San_Marino = "Europe/San_Marino"
    Europe_Sarajevo = "Europe/Sarajevo"
    Europe_Saratov = "Europe/Saratov"
    Europe_Simferopol = "Europe/Simferopol"
    Europe_Skopje = "Europe/Skopje"
    Europe_Sofia = "Europe/Sofia"
    Europe_Stockholm = "Europe/Stockholm"
    Europe_Tallinn = "Europe/Tallinn"
    Europe_Tirane = "Europe/Tirane"
    Europe_Tiraspol = "Europe/Tiraspol"
    Europe_Ulyanovsk = "Europe/Ulyanovsk"
    Europe_Uzhgorod = "Europe/Uzhgorod"
    Europe_Vaduz = "Europe/Vaduz"
    Europe_Vatican = "Europe/Vatican"
    Europe_Vienna = "Europe/Vienna"
    Europe_Vilnius = "Europe/Vilnius"
    Europe_Volgograd = "Europe/Volgograd"
    Europe_Warsaw = "Europe/Warsaw"
    Europe_Zagreb = "Europe/Zagreb"
    Europe_Zaporozhye = "Europe/Zaporozhye"
    Europe_Zurich = "Europe/Zurich"
    Factory = "Factory"
    GB = "GB"
    GB_Eire = "GB-Eire"
    GMT = "GMT"
    GMT_0 = "GMT+0"
    GMT_0_1 = "GMT-0"
    GMT0 = "GMT0"
    Greenwich = "Greenwich"
    HST = "HST"
    Hongkong = "Hongkong"
    Iceland = "Iceland"
    Indian_Antananarivo = "Indian/Antananarivo"
    Indian_Chagos = "Indian/Chagos"
    Indian_Christmas = "Indian/Christmas"
    Indian_Cocos = "Indian/Cocos"
    Indian_Comoro = "Indian/Comoro"
    Indian_Kerguelen = "Indian/Kerguelen"
    Indian_Mahe = "Indian/Mahe"
    Indian_Maldives = "Indian/Maldives"
    Indian_Mauritius = "Indian/Mauritius"
    Indian_Mayotte = "Indian/Mayotte"
    Indian_Reunion = "Indian/Reunion"
    Iran = "Iran"
    Israel = "Israel"
    Jamaica = "Jamaica"
    Japan = "Japan"
    Kwajalein = "Kwajalein"
    Libya = "Libya"
    MET = "MET"
    MST = "MST"
    MST7MDT = "MST7MDT"
    Mexico_BajaNorte = "Mexico/BajaNorte"
    Mexico_BajaSur = "Mexico/BajaSur"
    Mexico_General = "Mexico/General"
    NZ = "NZ"
    NZ_CHAT = "NZ-CHAT"
    Navajo = "Navajo"
    PRC = "PRC"
    PST8PDT = "PST8PDT"
    Pacific_Apia = "Pacific/Apia"
    Pacific_Auckland = "Pacific/Auckland"
    Pacific_Bougainville = "Pacific/Bougainville"
    Pacific_Chatham = "Pacific/Chatham"
    Pacific_Chuuk = "Pacific/Chuuk"
    Pacific_Easter = "Pacific/Easter"
    Pacific_Efate = "Pacific/Efate"
    Pacific_Enderbury = "Pacific/Enderbury"
    Pacific_Fakaofo = "Pacific/Fakaofo"
    Pacific_Fiji = "Pacific/Fiji"
    Pacific_Funafuti = "Pacific/Funafuti"
    Pacific_Galapagos = "Pacific/Galapagos"
    Pacific_Gambier = "Pacific/Gambier"
    Pacific_Guadalcanal = "Pacific/Guadalcanal"
    Pacific_Guam = "Pacific/Guam"
    Pacific_Honolulu = "Pacific/Honolulu"
    Pacific_Johnston = "Pacific/Johnston"
    Pacific_Kanton = "Pacific/Kanton"
    Pacific_Kiritimati = "Pacific/Kiritimati"
    Pacific_Kosrae = "Pacific/Kosrae"
    Pacific_Kwajalein = "Pacific/Kwajalein"
    Pacific_Majuro = "Pacific/Majuro"
    Pacific_Marquesas = "Pacific/Marquesas"
    Pacific_Midway = "Pacific/Midway"
    Pacific_Nauru = "Pacific/Nauru"
    Pacific_Niue = "Pacific/Niue"
    Pacific_Norfolk = "Pacific/Norfolk"
    Pacific_Noumea = "Pacific/Noumea"
    Pacific_Pago_Pago = "Pacific/Pago_Pago"
    Pacific_Palau = "Pacific/Palau"
    Pacific_Pitcairn = "Pacific/Pitcairn"
    Pacific_Pohnpei = "Pacific/Pohnpei"
    Pacific_Ponape = "Pacific/Ponape"
    Pacific_Port_Moresby = "Pacific/Port_Moresby"
    Pacific_Rarotonga = "Pacific/Rarotonga"
    Pacific_Saipan = "Pacific/Saipan"
    Pacific_Samoa = "Pacific/Samoa"
    Pacific_Tahiti = "Pacific/Tahiti"
    Pacific_Tarawa = "Pacific/Tarawa"
    Pacific_Tongatapu = "Pacific/Tongatapu"
    Pacific_Truk = "Pacific/Truk"
    Pacific_Wake = "Pacific/Wake"
    Pacific_Wallis = "Pacific/Wallis"
    Pacific_Yap = "Pacific/Yap"
    Poland = "Poland"
    Portugal = "Portugal"
    ROC = "ROC"
    ROK = "ROK"
    Singapore = "Singapore"
    Turkey = "Turkey"
    UCT = "UCT"
    US_Alaska = "US/Alaska"
    US_Aleutian = "US/Aleutian"
    US_Arizona = "US/Arizona"
    US_Central = "US/Central"
    US_East_Indiana = "US/East-Indiana"
    US_Eastern = "US/Eastern"
    US_Hawaii = "US/Hawaii"
    US_Indiana_Starke = "US/Indiana-Starke"
    US_Michigan = "US/Michigan"
    US_Mountain = "US/Mountain"
    US_Pacific = "US/Pacific"
    US_Samoa = "US/Samoa"
    UTC = "UTC"
    Universal = "Universal"
    W_SU = "W-SU"
    WET = "WET"
    Zulu = "Zulu"


@dataclass
class TimezoneRequest(BaseModel):
    timezone: TimezoneEnum


@dataclass
class User(BaseModel):
    username: str


@dataclass
class UserProfile(BaseModel):
    user: User
    mugshot: str
    institution: str
    department: str
    location: LocationEnum
    website: str | None


@dataclass
class UserRequest(BaseModel):
    username: str


@dataclass
class UserUpload(BaseModel):
    pk: str
    created: str
    filename: str
    key: str
    s3_upload_id: str
    status: str
    api_url: str


@dataclass
class UserUploadComplete(BaseModel):
    pk: str
    created: str
    filename: str
    key: str
    s3_upload_id: str
    status: str
    api_url: str


@dataclass
class UserUploadCreate(BaseModel):
    pk: str
    created: str
    filename: str
    key: str
    s3_upload_id: str
    status: str
    api_url: str


@dataclass
class UserUploadCreateRequest(BaseModel):
    filename: str


@dataclass
class UserUploadParts(BaseModel):
    pk: str
    created: str
    filename: str
    key: str
    s3_upload_id: str
    status: str
    api_url: str
    parts: list[Part]


@dataclass
class UserUploadPresignedURLs(BaseModel):
    pk: str
    created: str
    filename: str
    key: str
    s3_upload_id: str
    status: str
    api_url: str
    presigned_urls: dict[str, str]


@dataclass
class WindowPreset(BaseModel):
    pk: int
    slug: str
    title: str
    description: str | None
    center: int | None
    width: int | None
    lower_percentile: int | None
    upper_percentile: int | None


@dataclass
class Workstation(BaseModel):
    pk: str
    title: str
    slug: str
    url: str


@dataclass
class WorkstationConfig(BaseModel):
    pk: str
    slug: str
    title: str
    description: str | None
    created: str
    modified: str
    creator: str
    image_context: str
    window_presets: list[WindowPreset]
    default_window_preset: WindowPreset
    default_slab_thickness_mm: float
    default_slab_render_method: str
    default_orientation: str
    default_overlay_alpha: float
    overlay_luts: list[LookUpTable]
    default_overlay_lut: LookUpTable
    default_overlay_interpolation: str
    default_image_interpolation: str
    default_limit_view_area_to_image_volume: bool | None
    ghosting_slice_depth: int | None
    overlay_segments: Any | None
    key_bindings: Any | None
    default_zoom_scale: float
    default_brush_size: Decimal | None
    default_annotation_color: str | None
    default_annotation_line_width: int | None
    show_image_info_plugin: bool | None
    show_display_plugin: bool | None
    show_image_switcher_plugin: bool | None
    show_algorithm_output_plugin: bool | None
    show_overlay_plugin: bool | None
    show_annotation_statistics_plugin: bool | None
    show_swivel_tool: bool | None
    show_invert_tool: bool | None
    show_flip_tool: bool | None
    show_window_level_tool: bool | None
    show_reset_tool: bool | None
    show_overlay_selection_tool: bool | None
    show_lut_selection_tool: bool | None
    show_annotation_counter_tool: bool | None
    enabled_preprocessors: list[str]
    auto_jump_center_of_gravity: bool | None
    link_images: bool | None
    link_panning: bool | None
    link_zooming: bool | None
    link_slicing: bool | None
    link_orienting: bool | None
    link_windowing: bool | None
    link_inverting: bool | None
    link_flipping: bool | None
    point_bounding_box_size_mm: float


@dataclass
class ArchiveItemPost(BaseModel):
    pk: str
    archive: str | None
    values: list[ComponentInterfaceValuePost] | None
    hanging_protocol: HangingProtocol | None
    optional_hanging_protocols: list[HangingProtocol]
    view_content: Any
    title: str = ""


@dataclass
class ArchiveItemPostRequest(BaseModel):
    archive: str | None
    values: list[ComponentInterfaceValuePostRequest] | None
    title: str = ""


@dataclass
class ComponentInterface(BaseModel):
    title: str
    description: str | None
    slug: str
    kind: str
    pk: int
    default_value: Any | None
    super_kind: str
    relative_path: str
    overlay_segments: Any | None
    look_up_table: LookUpTable | None


@dataclass
class ComponentInterfaceValue(BaseModel):
    interface: ComponentInterface
    value: Any | None
    file: str | None
    image: SimpleImage | None
    pk: int


@dataclass
class DisplaySetPost(BaseModel):
    pk: str
    title_safe: str
    reader_study: str | None
    values: list[ComponentInterfaceValuePost] | None
    order: int | None
    api_url: str
    hanging_protocol: HangingProtocol | None
    optional_hanging_protocols: list[HangingProtocol]
    view_content: Any
    description: str
    index: int | None
    title: str = ""


@dataclass
class ExternalEvaluationUpdate(BaseModel):
    metrics: Any | None
    status: StatusEnum
    error_message: str | None


@dataclass
class HyperlinkedComponentInterfaceValue(BaseModel):
    interface: ComponentInterface
    value: Any | None
    file: str | None
    image: str | None
    pk: int


@dataclass
class HyperlinkedJob(BaseModel):
    pk: str
    url: str
    api_url: str
    algorithm_image: str
    inputs: list[HyperlinkedComponentInterfaceValue]
    outputs: list[HyperlinkedComponentInterfaceValue]
    status: str
    hanging_protocol: HangingProtocol | None
    optional_hanging_protocols: list[HangingProtocol]
    view_content: Any
    exec_duration: str
    invoke_duration: str
    algorithm: str


@dataclass
class ImageFile(BaseModel):
    pk: str
    image: str | None
    file: str
    image_type: ImageTypeEnum | None


@dataclass
class PaginatedComponentInterfaceList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[ComponentInterface]


@dataclass
class PaginatedHyperlinkedJobList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[HyperlinkedJob]


@dataclass
class PaginatedPublicChallengeList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[PublicChallenge]


@dataclass
class PaginatedRawImageUploadSessionList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[RawImageUploadSession]


@dataclass
class PaginatedSessionList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Session]


@dataclass
class PaginatedUserUploadList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[UserUpload]


@dataclass
class PaginatedWorkstationConfigList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[WorkstationConfig]


@dataclass
class PaginatedWorkstationList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Workstation]


@dataclass
class PatchedExternalEvaluationUpdateRequest(BaseModel):
    metrics: Any | None
    status: StatusEnum | None
    error_message: str | None


@dataclass
class Question(BaseModel):
    answer_type: str
    api_url: str
    form_direction: str
    help_text: str
    help_text_safe: str
    image_port: str
    default_annotation_color: str | None
    pk: str
    question_text: str
    question_text_safe: str
    reader_study: str
    required: bool | None
    options: list[CategoricalOption]
    interface: ComponentInterface | None
    overlay_segments: Any | None
    look_up_table: LookUpTable | None
    widget: str
    answer_min_value: int | None
    answer_max_value: int | None
    answer_step_size: Decimal | None
    answer_min_length: int | None
    answer_max_length: int | None
    answer_match_pattern: str | None
    empty_answer_confirmation: bool | None
    empty_answer_confirmation_label: str
    empty_answer_confirmation_label_safe: str
    interactive_algorithms: list[str]


@dataclass
class ReaderStudy(BaseModel):
    api_url: str
    url: str
    slug: str
    logo: str
    description: str | None
    help_text: str
    help_text_safe: str
    pk: str
    questions: list[Question]
    title: str
    title_safe: str
    is_educational: bool | None
    instant_verification: bool | None
    has_ground_truth: bool
    allow_answer_modification: bool | None
    enable_autosaving: bool | None
    allow_case_navigation: bool | None
    allow_show_all_annotations: bool | None
    roll_over_answers_for_n_cases: int | None
    end_of_study_text_safe: str


@dataclass
class Submission(BaseModel):
    pk: str
    phase: Phase
    created: str
    creator: User
    comment: str | None
    predictions_file: str | None
    supplementary_file: str | None
    supplementary_url: str | None
    algorithm_image: str


@dataclass
class SubmissionRequest(BaseModel):
    phase: PhaseRequest
    creator: UserRequest
    comment: str | None
    predictions_file: bytes | None
    supplementary_file: bytes | None
    supplementary_url: str | None


@dataclass
class Timezone(BaseModel):
    timezone: TimezoneEnum


@dataclass
class AlgorithmInterface(BaseModel):
    inputs: list[ComponentInterface]
    outputs: list[ComponentInterface]


@dataclass
class ArchiveItem(BaseModel):
    pk: str
    archive: str
    values: list[HyperlinkedComponentInterfaceValue]
    hanging_protocol: HangingProtocol | None
    optional_hanging_protocols: list[HangingProtocol]
    view_content: Any
    title: str = ""


@dataclass
class DisplaySet(BaseModel):
    pk: str
    title_safe: str
    reader_study: str
    values: list[HyperlinkedComponentInterfaceValue]
    order: int | None
    api_url: str
    hanging_protocol: HangingProtocol | None
    optional_hanging_protocols: list[HangingProtocol]
    view_content: Any
    description: str
    index: int | None
    title: str = ""


@dataclass
class Evaluation(BaseModel):
    pk: str
    method: str | None
    submission: Submission
    created: str
    published: bool | None
    outputs: list[ComponentInterfaceValue]
    rank: int | None
    rank_score: float | None
    rank_per_metric: Any | None
    status: str
    title: str
    exec_duration: str
    invoke_duration: str


@dataclass
class ExternalEvaluation(BaseModel):
    pk: str
    method: str | None
    submission: Submission
    created: str
    published: bool | None
    outputs: list[ComponentInterfaceValue]
    rank: int | None
    rank_score: float | None
    rank_per_metric: Any | None
    status: str
    title: str
    exec_duration: str
    invoke_duration: str
    algorithm_model: dict[str, Any] | None
    algorithm_image: dict[str, Any]
    claimed_by: int | None
    phase_pk: str


@dataclass
class HyperlinkedImage(BaseModel):
    pk: str
    name: str
    files: list[ImageFile]
    dicom_image_set: DICOMImageSet | None
    width: int | None
    height: int | None
    depth: int | None
    color_space: ColorSpaceEnum | BlankEnum | None
    modality: ImagingModality | None
    eye_choice: EyeChoiceEnum | None
    stereoscopic_choice: StereoscopicChoiceEnum | BlankEnum | NullEnum | None
    field_of_view: FieldOfViewEnum | BlankEnum | NullEnum | None
    shape_without_color: list[int]
    shape: list[int]
    voxel_width_mm: float | None
    voxel_height_mm: float | None
    voxel_depth_mm: float | None
    api_url: str
    patient_id: str | None
    patient_name: str | None
    patient_birth_date: str | None
    patient_age: str | None
    patient_sex: PatientSexEnum | BlankEnum | None
    study_date: str | None
    study_instance_uid: str | None
    series_instance_uid: str | None
    study_description: str | None
    series_description: str | None
    window_center: float | None
    window_width: float | None
    segments: Any | None


@dataclass
class PaginatedArchiveItemList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[ArchiveItem]


@dataclass
class PaginatedDisplaySetList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[DisplaySet]


@dataclass
class PaginatedEvaluationList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Evaluation]


@dataclass
class PaginatedHyperlinkedImageList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[HyperlinkedImage]


@dataclass
class PaginatedQuestionList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Question]


@dataclass
class PaginatedReaderStudyList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[ReaderStudy]


@dataclass
class PatchedExternalEvaluationRequest(BaseModel):
    method: str | None
    submission: SubmissionRequest | None
    published: bool | None
    rank: int | None
    rank_score: float | None
    rank_per_metric: Any | None
    claimed_by: int | None
    phase_pk: str | None


@dataclass
class Algorithm(BaseModel):
    api_url: str
    url: str
    description: str | None
    pk: str
    title: str
    logo: str
    slug: str
    average_duration: float | None
    interfaces: list[AlgorithmInterface]


@dataclass
class PaginatedAlgorithmList(BaseModel):
    count: int
    next: str | None
    previous: str | None
    results: list[Algorithm]



import json
import logging
import os
import re
import sys
import uuid
from asyncio import Semaphore
from collections.abc import Callable, Iterator
from contextlib import ExitStack
from io import BytesIO
from pathlib import Path
from random import randint
from time import sleep
from typing import IO, Any, cast, get_type_hints
from urllib.parse import urljoin

import httpx
from asgiref.sync import async_to_sync, sync_to_async
from httpx import URL, HTTPStatusError, Timeout

if sys.version_info >= (3, 11):
    from asyncio import TaskGroup
else:  # Use backport for Python <3.11
    from taskgroup import TaskGroup

import gcapi.models
from gcapi.apibase import APIBase, ModifiableMixin
from gcapi.check_version import check_version
from gcapi.create_strategies import (
    JobInputsCreateStrategy,
    SocketValueCreateStrategy,
    SocketValueSpec,
    select_socket_value_strategy,
)
from gcapi.exceptions import ObjectNotFound, SocketNotFound
from gcapi.retries import BaseRetryStrategy, SelectiveBackoffStrategy
from gcapi.transports import RetryTransport
from gcapi.typing import SocketValuePostSet

logger = logging.getLogger(__name__)


def is_uuid(s: str) -> bool:
    """
    Check if a string is a valid UUID.

    Args:
        s: The string to check for UUID validity.

    Returns:
        True if the string is a valid UUID, False otherwise.
    """
    try:
        uuid.UUID(s)
    except ValueError:
        return False
    else:
        return True


class ImagesAPI(APIBase[gcapi.models.HyperlinkedImage]):
    base_path = "cases/images/"
    model = gcapi.models.HyperlinkedImage

    def download(
        self,
        *,
        output_directory: str | Path,
        filename: str | None = None,
        image_type: str | None = None,
        pk: str | None = None,
        url: str | None = None,
        files: list | None = None,
        **params: Any,
    ) -> list[Path]:
        """
        Download image files to local filesystem.

        Args:
            output_directory: Directory to save downloaded files.
            filename: Optional, base filename for downloaded files. Extension is added automatically.
            image_type: Restrict download to a particular image type.
            pk: Primary key of the image to download.
            url: API URL of the image to download.
            files: List of file objects to download directly.
            **params: Additional parameters for image detail lookup.

        Returns:
            List of Path objects for downloaded files.

        Raises:
            ValueError: If not exactly one of pk, url, files, or params is specified.
        """
        if len([p for p in (pk, url, files, params) if p]) != 1:
            raise ValueError(
                "Exactly one of pk, url, files or params must be specified"
            )

        # Retrieve details of the image if needed
        if files is None:
            if pk is not None:
                image = self.detail(pk=pk)
            elif url is not None:
                image = self.detail(api_url=url)
            else:
                image = self.detail(**params)

            if image.dicom_image_set is not None:
                if filename is not None:
                    raise ValueError(
                        "Do not provide a filename for DICOM image sets"
                    )

                return self._download_dicom_image_set(
                    pk=image.pk,
                    output_directory=output_directory,
                )

            filename = filename or image.pk
            files = image.files

        # Make sure destination exists
        output_directory = Path(output_directory).absolute()
        output_directory.mkdir(parents=True, exist_ok=True)

        filename = filename or "image"

        # Download the files
        downloaded_files = []
        for file in files:
            if image_type and file.image_type != image_type:
                continue

            data = self._client(
                method="GET", url=file.file, follow_redirects=True
            ).content

            suffix = file.file.split(".")[-1]
            local_file = output_directory / f"{filename}.{suffix}"

            if local_file.exists():
                raise FileExistsError(f"File {local_file} already exists")

            with local_file.open("wb") as fp:
                fp.write(data)

            downloaded_files.append(local_file)

        return downloaded_files

    @async_to_sync
    async def _download_dicom_image_set(
        self,
        *,
        pk: str,
        output_directory: Path | str,
    ) -> list[Path]:
        """
        Download all DICOM instances of a DICOM image set.

        Args:
            pk: Primary key of the image to download.
            output_directory: Directory to save downloaded files.
            filename: Optional prefix filename for downloaded files.

        Returns:
            List of Path objects for downloaded DICOM files.
        """
        resp = self._client(
            path=f"cases/images/{pk}/dicom/instances/",
            method="GET",
        )

        output = Path(output_directory).absolute()
        output.mkdir(parents=True, exist_ok=True)

        semaphore = Semaphore(self._client.max_concurrent_downloads)

        async def download_with_semaphore(instance):
            sop_instance_uid = instance["sop_instance_uid"]
            output_file = output / f"{sop_instance_uid}.dcm"

            if output_file.exists():
                raise FileExistsError(f"File {output_file} already exists")

            async with semaphore:
                return await self._download_dicom_instance(
                    stream_kwargs=instance["get_instance"],
                    file=output_file,
                )

        tasks = []
        async with TaskGroup() as tg:
            for instance in resp["instances"]:
                task = tg.create_task(download_with_semaphore(instance))
                tasks.append(task)

        return [t.result() for t in tasks]

    @sync_to_async(thread_sensitive=False)
    def _download_dicom_instance(
        self, stream_kwargs: dict[str, Any], file: Path
    ):

        with open(file, "wb") as f:
            with httpx.stream(
                **stream_kwargs,
            ) as resp:
                resp.raise_for_status()
                for chunk in resp.iter_bytes():
                    f.write(chunk)
        return file


class UploadSessionsAPI(
    APIBase[gcapi.models.RawImageUploadSession],
):
    base_path = "cases/upload-sessions/"
    model = gcapi.models.RawImageUploadSession
    response_model = gcapi.models.RawImageUploadSession


class WorkstationSessionsAPI(APIBase[gcapi.models.Session]):
    base_path = "workstations/sessions/"
    model = gcapi.models.Session
    response_model = gcapi.models.RawImageUploadSession


class ReaderStudyQuestionsAPI(APIBase[gcapi.models.Question]):
    base_path = "reader-studies/questions/"
    model = gcapi.models.Question


class ReaderStudyMineAnswersAPI(
    ModifiableMixin[gcapi.models.Answer],
    APIBase[gcapi.models.Answer],
):
    base_path = "reader-studies/answers/mine/"
    model = gcapi.models.Answer
    response_model = gcapi.models.Answer


class ReaderStudyAnswersAPI(
    ModifiableMixin[gcapi.models.Answer],
    APIBase[gcapi.models.Answer],
):
    base_path = "reader-studies/answers/"
    model = gcapi.models.Answer
    response_model = gcapi.models.Answer

    sub_apis = {"mine": ReaderStudyMineAnswersAPI}

    mine = None  # type: ReaderStudyMineAnswersAPI

    def _process_request_arguments(self, data):
        if data is not None:
            key_and_url = {
                "question": ReaderStudyQuestionsAPI.base_path,
                "display_set": ReaderStudyDisplaySetsAPI.base_path,
            }
            for key, api in key_and_url.items():
                if is_uuid(data.get(key, "")):
                    data[key] = str(
                        self._client.base_url.join(api).join(data[key] + "/")
                    )

        return super()._process_request_arguments(data)


class ReaderStudyDisplaySetsAPI(
    ModifiableMixin[gcapi.models.DisplaySetPost],
    APIBase[gcapi.models.DisplaySet],
):
    base_path = "reader-studies/display-sets/"
    model = gcapi.models.DisplaySet
    response_model = gcapi.models.DisplaySetPost


class ReaderStudiesAPI(APIBase[gcapi.models.ReaderStudy]):
    base_path = "reader-studies/"
    model = gcapi.models.ReaderStudy

    sub_apis = {
        "answers": ReaderStudyAnswersAPI,
        "questions": ReaderStudyQuestionsAPI,
        "display_sets": ReaderStudyDisplaySetsAPI,
    }

    answers = None  # type: ReaderStudyAnswersAPI
    questions = None  # type: ReaderStudyQuestionsAPI
    display_sets = None  # type: ReaderStudyDisplaySetsAPI

    def ground_truth(self, pk: str, case_pk: str) -> dict:
        """
        Get ground truth data for a specific case in a reader study.

        Args:
            pk: Primary key of the reader study.
            case_pk: Primary key of the case.

        Returns:
            Ground truth data for the specified case.
        """
        return self._client(
            method="GET",
            path=urljoin(
                self.base_path, pk + "/ground-truth/" + case_pk + "/"
            ),
        )


class AlgorithmsAPI(APIBase[gcapi.models.Algorithm]):
    base_path = "algorithms/"
    model = gcapi.models.Algorithm


class AlgorithmJobsAPI(
    ModifiableMixin[gcapi.models.JobPost],
    APIBase[gcapi.models.HyperlinkedJob],
):
    base_path = "algorithms/jobs/"
    model = gcapi.models.HyperlinkedJob
    response_model = gcapi.models.JobPost

    def by_input_image(self, pk: str) -> Iterator[gcapi.models.HyperlinkedJob]:
        """
        Get algorithm jobs filtered by input image.

        Args:
            pk: Primary key of the input image to filter jobs by.

        Yields:
            Algorithm job instances that use the specified input image.
        """
        yield from self.iterate_all(params={"image": pk})


class AlgorithmImagesAPI(APIBase[gcapi.models.AlgorithmImage]):
    base_path = "algorithms/images/"
    model = gcapi.models.AlgorithmImage


class ArchivesAPI(APIBase[gcapi.models.Archive]):
    base_path = "archives/"
    model = gcapi.models.Archive


class ArchiveItemsAPI(
    ModifiableMixin[gcapi.models.ArchiveItemPost],
    APIBase[gcapi.models.ArchiveItem],
):
    base_path = "archives/items/"
    model = gcapi.models.ArchiveItem
    response_model = gcapi.models.ArchiveItemPost


class ComponentInterfacesAPI(APIBase[gcapi.models.ComponentInterface]):
    base_path = "components/interfaces/"
    model = gcapi.models.ComponentInterface


class UploadsAPI(APIBase[gcapi.models.UserUpload]):
    base_path = "uploads/"
    model = gcapi.models.UserUpload

    chunk_size = 32 * 1024 * 1024
    n_presigned_urls = 5  # number of pre-signed urls to generate
    max_retries = 10

    def create(self, *, filename: str | Path) -> gcapi.models.UserUpload:
        """
        Create a new upload session.

        Args:
            filename: Name of the file to be uploaded.

        Returns:
            The created upload session model instance.
        """
        result = self._client(
            method="POST",
            path=self.base_path,
            json={"filename": str(filename)},
        )
        return self.model(**result)

    def generate_presigned_urls(
        self, *, pk: str, s3_upload_id: str, part_numbers: list[int]
    ) -> dict[str, str]:
        """
        Generate presigned URLs for multipart upload parts.

        Args:
            pk: Primary key of the upload session.
            s3_upload_id: S3 multipart upload identifier.
            part_numbers: List of part numbers to generate URLs for.

        Returns:
            Dictionary containing presigned URLs for each part number.
        """
        url = urljoin(
            self.base_path, f"{pk}/{s3_upload_id}/generate-presigned-urls/"
        )
        return self._client(
            method="PATCH", path=url, json={"part_numbers": part_numbers}
        )

    def abort_multipart_upload(self, *, pk: str, s3_upload_id: str) -> dict:
        """
        Abort a multipart upload session.

        Args:
            pk: Primary key of the upload session.
            s3_upload_id: S3 multipart upload identifier to abort.

        Returns:
            Response from the API.
        """
        url = urljoin(
            self.base_path, f"{pk}/{s3_upload_id}/abort-multipart-upload/"
        )
        return self._client(method="PATCH", path=url)

    def complete_multipart_upload(
        self, *, pk: str, s3_upload_id: str, parts: list
    ) -> dict:
        """
        Complete a multipart upload session.

        Args:
            pk: Primary key of the upload session.
            s3_upload_id: S3 multipart upload identifier.
            parts: List of completed parts with ETag and PartNumber.

        Returns:
            Response from the API containing upload completion details.
        """
        url = urljoin(
            self.base_path, f"{pk}/{s3_upload_id}/complete-multipart-upload/"
        )
        return self._client(method="PATCH", path=url, json={"parts": parts})

    def list_parts(self, *, pk: str, s3_upload_id: str) -> dict:
        """
        List parts of a multipart upload.

        Args:
            pk: Primary key of the upload session.
            s3_upload_id: S3 multipart upload identifier.

        Returns:
            Response containing list of uploaded parts.
        """
        url = urljoin(self.base_path, f"{pk}/{s3_upload_id}/list-parts/")
        return self._client(path=url)

    def upload_fileobj(
        self,
        *,
        fileobj: IO,
        filename: str,
    ) -> gcapi.models.UserUpload:
        """
        Upload a file object using multipart upload.

        Args:
            fileobj: File object to upload
            filename: Name of the file being uploaded.

        Returns:
            gcapi.models.UserUpload: The completed upload model instance.

        Raises:
            Exception: If upload fails, the multipart upload is aborted and exception is re-raised.
        """
        user_upload = self.create(filename=filename)

        pk = user_upload.pk
        s3_upload_id = user_upload.s3_upload_id

        try:
            parts = self._put_fileobj(
                fileobj=fileobj, pk=pk, s3_upload_id=s3_upload_id
            )
        except Exception:
            self.abort_multipart_upload(pk=pk, s3_upload_id=s3_upload_id)
            raise

        result = self.complete_multipart_upload(
            pk=pk, s3_upload_id=s3_upload_id, parts=parts
        )
        return self.model(**result)  # noqa: B901

    def _put_fileobj(self, *, fileobj, pk, s3_upload_id):
        part_number = 1  # s3 uses 1-indexed chunks
        presigned_urls = {}
        parts = []

        while True:
            chunk = fileobj.read(self.chunk_size)

            if not chunk:
                break

            if str(part_number) not in presigned_urls:
                presigned_urls.update(
                    self._get_next_presigned_urls(
                        pk=pk,
                        s3_upload_id=s3_upload_id,
                        part_number=part_number,
                    )
                )

            response = self._put_chunk(
                chunk=chunk, url=presigned_urls[str(part_number)]
            )

            parts.append(
                {"ETag": response.headers["ETag"], "PartNumber": part_number}
            )

            part_number += 1

        return parts

    def _get_next_presigned_urls(self, *, pk, s3_upload_id, part_number):
        response = self.generate_presigned_urls(
            pk=pk,
            s3_upload_id=s3_upload_id,
            part_numbers=[
                *range(part_number, part_number + self.n_presigned_urls)
            ],
        )
        return response["presigned_urls"]

    def _put_chunk(self, *, chunk, url):
        num_retries = 0
        e = Exception()

        if isinstance(chunk, BytesIO):
            chunk = chunk.read()

        while num_retries < self.max_retries:
            try:
                result = self._client.request(
                    method="PUT", url=url, content=chunk
                )
                break
            except HTTPStatusError as _e:
                status_code = _e.response.status_code
                if status_code in [409, 423] or status_code >= 500:
                    num_retries += 1
                    e = _e
                    sleep((2**num_retries) + (randint(0, 1000) / 1000))
                else:
                    raise
        else:
            raise e

        return result


class WorkstationConfigsAPI(APIBase[gcapi.models.WorkstationConfig]):
    base_path = "workstations/configs/"
    model = gcapi.models.WorkstationConfig


def _generate_auth_header(token: str = "") -> dict:
    """
    Generate authorization header for API requests.

    Args:
        token (str, optional): Authorization token. If empty, retrieves token from
        GRAND_CHALLENGE_AUTHORIZATION environment variable.

    Returns:
        dict: Dictionary containing the Authorization header.

    Raises:
        RuntimeError: If no token is provided and GRAND_CHALLENGE_AUTHORIZATION
        environment variable is not set, or if token format is invalid.
    """
    if not token:
        try:
            token = str(os.environ["GRAND_CHALLENGE_AUTHORIZATION"])
        except KeyError:
            raise RuntimeError("Token must be set") from None

    token = re.sub(" +", " ", token)
    token_parts = token.strip().split(" ")

    if len(token_parts) not in [1, 2]:
        raise RuntimeError("Invalid token format")

    return {"Authorization": f"BEARER {token_parts[-1]}"}


class ApiDefinitions:
    images: ImagesAPI
    reader_studies: ReaderStudiesAPI
    sessions: WorkstationSessionsAPI
    uploads: UploadsAPI
    algorithms: AlgorithmsAPI
    algorithm_jobs: AlgorithmJobsAPI
    algorithm_images: AlgorithmImagesAPI
    archives: ArchivesAPI
    workstation_configs: WorkstationConfigsAPI
    raw_image_upload_sessions: UploadSessionsAPI
    archive_items: ArchiveItemsAPI
    interfaces: ComponentInterfacesAPI


class Client(httpx.Client, ApiDefinitions):
    """The Grand Challenge API client."""

    _api_meta: ApiDefinitions

    def __init__(
        self,
        token: str = "",
        base_url: str = "https://grand-challenge.org/api/v1/",
        verify: bool = True,
        timeout: float = 60.0,
        retry_strategy: Callable[[], BaseRetryStrategy] | None = None,
        max_concurrent_downloads: int = 10,
    ):
        """
        Args:
            token: Authorization token for API access. If not provided, will be read
                from GRAND_CHALLENGE_AUTHORIZATION environment variable.
            base_url: Base URL for the API.
            verify: Whether to verify SSL certificates.
            timeout: Request timeout in seconds.
            retry_strategy: Factory function that returns a retry strategy instance. If None,
                uses SelectiveBackoffStrategy with default parameters.
            max_concurrent_downloads: Maximum number of concurrent downloads allowed.
        """
        check_version(base_url=base_url)

        retry_strategy = retry_strategy or SelectiveBackoffStrategy(
            backoff_factor=0.1,
            maximum_number_of_retries=8,  # ~25.5 seconds total backoff
        )
        httpx.Client.__init__(
            self,
            verify=verify,
            timeout=Timeout(timeout=timeout),
            transport=RetryTransport(
                verify=verify,
                retry_strategy=retry_strategy,
            ),
        )

        self.headers.update({"Accept": "application/json"})
        self._auth_header = _generate_auth_header(token=token)

        self.base_url = URL(base_url)
        if self.base_url.scheme.lower() != "https":
            raise RuntimeError("Base URL must be https")

        self.max_concurrent_downloads = max_concurrent_downloads

        self._api_meta = ApiDefinitions()
        for name, cls in get_type_hints(ApiDefinitions).items():
            setattr(self._api_meta, name, cls(client=self))

        self._socket_cache: dict[str, gcapi.models.ComponentInterface] = {}
        self._algorithm_cache: dict[str, gcapi.models.Algorithm] = {}
        self._archive_cache: dict[str, gcapi.models.Archive] = {}

    def __getattr__(self, item):
        api = getattr(self._api_meta, item, None)
        if api:
            return api
        else:
            raise AttributeError(f"Client has no function or API {item!r}")

    def _validate_url(self, url):
        url = URL(url)

        if not url.scheme == "https" or url.netloc != self.base_url.netloc:
            raise RuntimeError(f"Invalid target URL: {url}")

    def __call__(
        self,
        method="GET",
        url="",
        path="",
        params=None,
        json=None,
        extra_headers=None,
        files=None,
        data=None,
        follow_redirects=False,
    ) -> Any:
        """
        Make an HTTP request to the API.

        Args:
            method (str, optional): HTTP method to use, by default "GET".
            url (str, optional): Full URL to request. If provided, path is ignored.
            path (str, optional): Path relative to base_url. Ignored if url is provided.
            params (dict, optional): Query parameters to include in the request.
            json (dict, optional): JSON data to send in the request body.
            extra_headers (dict, optional): Additional headers to include in the request.
            files (dict, optional): Files to upload with the request.
            data (dict, optional): Form data to send in the request body.
            follow_redirects (bool, optional): Whether to follow HTTP redirects, by default False.

        Returns:
            Any: JSON response data if Content-Type is application/json,
                otherwise the raw response object.

        Raises:
            HTTPStatusError: If the HTTP request fails with a non-2xx status code.
        """
        if url:
            url = URL(url)
        else:
            url = self.base_url.join(path)
        if extra_headers is None:
            extra_headers = {}
        if json is not None:
            extra_headers["Content-Type"] = "application/json"

        self._validate_url(url)

        response = self.request(
            method=method,
            url=str(url),
            files={} if files is None else files,
            data={} if data is None else data,
            headers={
                **self.headers,
                **self._auth_header,
                **extra_headers,
            },
            params={} if params is None else params,
            json=json,
            follow_redirects=follow_redirects,
        )

        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            if e.response.headers.get("Content-Type") == "application/json":
                message = e.response.json()
                logger.error(f"{method} request to {url} failed: {message}")
            raise

        if response.headers.get("Content-Type") == "application/json":
            return response.json()
        else:
            return response

    def _fetch_algorithm_detail(self, slug: str) -> gcapi.models.Algorithm:
        if slug not in self._algorithm_cache:
            self._algorithm_cache[slug] = self.algorithms.detail(slug=slug)
        return self._algorithm_cache[slug]

    def download_socket_value(
        self,
        value: gcapi.models.HyperlinkedComponentInterfaceValue,
        *,
        output_directory: Path,
    ):
        """Download a socket value to the specified output directory.

        Uses the socket's relative path to determine the filename. The relative
        path is fixed and unique per socket, so downloading multiple different values
        will not overwrite each other.

        In addition, if the file to be downloaded already exists in the output
        directory, a FileExistsError is raised to prevent overwriting existing files.

        Args:
            value: The socket value to download.
            output_directory: The directory to download the value into.
        """

        target_path = output_directory / value.interface.relative_path
        target_path.parent.mkdir(parents=True, exist_ok=True)

        super_kind = value.interface.super_kind.casefold()
        if super_kind == "image":
            # Image values
            self.images.download(
                url=str(value.image),
                output_directory=target_path,
            )
        elif super_kind == "value":
            if target_path.exists():
                raise FileExistsError(f"File {target_path} already exists")
            # Direct values (e.g. '42')
            with open(target_path, "w") as f:
                json.dump(value.value, f, indent=2)
        elif super_kind == "file":
            if target_path.exists():
                raise FileExistsError(f"File {target_path} already exists")
            # Values stored as files
            resp = self(
                url=str(value.file),
                follow_redirects=True,
            )
            content = (
                resp.content
                if isinstance(resp, httpx.Response)
                else json.dumps(resp).encode()
            )
            with open(target_path, "wb") as f:
                f.write(content)
        else:
            raise ValueError(
                f"Unexpected super_kind {value.interface.super_kind}"
            )

    def start_algorithm_job(
        self,
        *,
        algorithm_slug: str,
        inputs: list[SocketValueSpec],
    ) -> gcapi.models.JobPost:
        """
        Starts an algorithm job with the provided inputs.

        ??? tip "Getting the interfaces of an algorithm"
            You can get the interfaces (i.e. all possible socket sets) of
            an algorithm by calling, and inspecting the .interface of the
            result of:

            ```Python
            client.algorithms.detail(slug="corads-ai")
            ```

        ??? tip "Re-using existing images"
            Existing images on Grand Challenge can be re-used by either
            passing an API url, or an existing socket value:

            ```Python
            from gcapi import SocketValueSpec

            image = client.images.detail(pk="ad5...")
            ds = client.reader_studies.display_sets.detail(pk="f5...")
            socket_value = ds.values[0]

            inputs = [
                SocketValueSpec(socket_slug="slug-0", existing_image_api_url=image.api_url),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=socket_value),
                SocketValueSpec(socket_slug="slug-2", existing_image_api_url=socket_value.image),
            ]
            ```

        ??? tip "Re-using existing socket values"
            Existing socket values from other display sets can be re-used by
            passing a socket value. The sockets must be the same.

            For instance:

            ```Python
            from gcapi import SocketValueSpec

            ds = client.reader_studies.display_sets.detail(pk="f5...")
            inputs = [
                SocketValueSpec(socket_slug="slug-0", existing_socket_value=ds.values[0]),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=ds.values[1]),
            ]
            ```

        Args:
            algorithm_slug: Slug for the algorithm (e.g. `"corads-ai"`).
                You can find this readily in the URL you use to visit the algorithm page:
                `https://grand-challenge.org/algorithms/corads-ai/`

            inputs: A list of socket value specifications.
                Each specification defines a socket slug and exactly one source
                (`value`, `file`, `files`, `existing_image_api_url`, or
                `existing_socket_value`).

        Returns:
            The newly created Job (post) object. Note that not all inputs will
                be immediately available until the background processing has
                completed.
        """

        algorithm = self._fetch_algorithm_detail(slug=algorithm_slug)

        with JobInputsCreateStrategy(
            algorithm=algorithm,
            inputs=inputs,
            client=self,
        ) as input_strategy:
            created_inputs = input_strategy()

            return self.algorithm_jobs.create(
                algorithm=algorithm.api_url,
                inputs=created_inputs,
            )

    def update_display_set(
        self, *, display_set_pk: str, values: list[SocketValueSpec]
    ) -> gcapi.models.DisplaySetPost:
        """
        This function patches an existing display set with the provided values.

        You can use this function, for example, to add metadata to a display set.

        If you provide a value or file for an existing socket of the display
        set, the old value will be overwritten by the new one, hence allowing you
        to update existing display-set values.

        ??? example

            First, retrieve the display sets from your reader study:

            ```Python
            reader_study = client.reader_studies.detail(slug="...")
            items = list(
                client.reader_studies.display_sets.iterate_all(
                    params={"reader_study": reader_study.pk}
                )
            )
            ```

            To then add, for example, a PDF report and a lung volume
            value to the first display set , provide the socket slugs together
            with the respective value or file path as follows:

            ```Python
            from gcapi import SocketValueSpec

            client.update_display_set(
                display_set_pk=items[0].pk,
                values=[
                    SocketValueSpec(socket_slug="report", file="report.pdf"),
                    SocketValueSpec(socket_slug="lung-volume", value=1.9),
                ],
            )
            ```

        Args:
            display_set_pk: The primary key of the display set to update.
            values: A list of socket value specifications.
                Each specification defines a socket slug and exactly one source
                (`value`, `file`, `files`, `existing_image_api_url`, or
                `existing_socket_value`).

        Returns:
            The updated display item (post) object. Note that not all values will
                be immediately available until the background processing has completed.
        """
        display_set = self._update_socket_value_set(
            target_pk=display_set_pk,
            values=values,
            api=self.reader_studies.display_sets,
        )
        return cast(gcapi.models.DisplaySetPost, display_set)

    def add_case_to_reader_study(
        self,
        *,
        reader_study_slug: str,
        values: list[SocketValueSpec],
    ) -> gcapi.models.DisplaySetPost:
        """
        This function takes a reader-study slug and a list of socket value specs.
        It then creates a single display set for the reader study.

        Example:
            ```Python
            from gcapi import SocketValueSpec

            client.add_case_to_reader_study(
                reader_study_slug="i-am-a-reader-study",
                values=[
                    SocketValueSpec(socket_slug="report", file="report.pdf"),
                    SocketValueSpec(socket_slug="lung-volume", value=1.9),
                ],
            )
            ```

        ??? tip "Re-using existing images"
            Existing images on Grand Challenge can be re-used by either
            passing an API url, or an existing socket value:

            ```Python
            from gcapi import SocketValueSpec


            image = client.images.detail(pk="ad5...")
            ds = client.reader_studies.display_sets.detail(pk="f5...")
            socket_value = ds.values[0]

            values = [
                SocketValueSpec(socket_slug="slug-0", existing_image_api_url=image.api_url),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=socket_value),
                SocketValueSpec(socket_slug="slug-2", existing_image_api_url=socket_value.image),
            ]
            ```

        ??? tip "Re-using existing socket values"
            Existing socket values from other display sets can be re-used by
            passing a socket value. The sockets must be the same.

            For instance:

            ```Python
            from gcapi import SocketValueSpec


            ds = client.reader_studies.display_sets.detail(pk="f5...")
            values = [
                SocketValueSpec(socket_slug="slug-0", existing_socket_value=ds.values[0]),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=ds.values[1]),
            ]
            ```

        Args:
            reader_study_slug: slug for the reader study (e.g. `"i-am-a-reader-study"`).
                You can find this readily in the URL you use to visit the reader-study page:
                `https://grand-challenge.org/reader-studies/i-am-a-reader-study/`

            values: A list of socket value specifications.
                Each specification defines a socket slug and exactly one source
                (`value`, `file`, `files`, `existing_image_api_url`, or
                `existing_socket_value`).

        Returns:
            The newly created display set (post) object. Note that not all values will
                be immediately available until the background processing has completed.
        """
        try:
            created_display_set = self._create_socket_value_set(
                creation_kwargs={"reader_study": reader_study_slug},
                values=values,
                api=self.reader_studies.display_sets,
            )
        except SocketNotFound as e:
            raise ValueError(
                f"{e.slug} is not an existing socket. "
                f"Please provide one from this list: "
                f"https://grand-challenge.org/components/interfaces/reader-studies/"
            ) from e

        return cast(gcapi.models.DisplaySetPost, created_display_set)

    def update_archive_item(
        self,
        *,
        archive_item_pk: str,
        values: list[SocketValueSpec],
    ) -> gcapi.models.ArchiveItemPost:
        """
        This function patches an existing archive item with the provided values.

        You can use this function, for example, to add metadata to an archive item.

        If you provide a value or file for an existing socket of the archive
        item, the old value will be overwritten by the new one, hence allowing you
        to update existing archive item values.

        ??? example
            First, retrieve the archive items from your archive:

            ```Python
            archive = client.archives.detail(slug="...")
            items = list(
                client.archive_items.iterate_all(params={"archive": archive.pk})
            )
            ```

            To then add, for example, a PDF report and a lung volume
            value to the first archive item , provide the socket slugs together
            with the respective value or file path as follows:

            ```Python
            from gcapi import SocketValueSpec

            client.update_archive_item(
                archive_item_pk=items[0].pk,
                values=[
                    SocketValueSpec(socket_slug="report", file="report.pdf"),
                    SocketValueSpec(socket_slug="lung-volume", value=1.9),
                ],
            )
            ```

        Args:
            archive_item_pk: The primary key of the archive item to update.
            values: A list of socket value specifications.
                Each specification defines a socket slug and exactly one source
                (`value`, `file`, `files`, `existing_image_api_url`, or
                `existing_socket_value`).

        Returns:
            The updated archive item (post) object. Note that not all values will
                be immediately available until the background processing has completed.
        """
        archive_item = self._update_socket_value_set(
            target_pk=archive_item_pk,
            values=values,
            api=self.archive_items,
        )
        return cast(gcapi.models.ArchiveItemPost, archive_item)

    def _fetch_archive_api_url(self, slug: str) -> str:
        if slug not in self._archive_cache:
            self._archive_cache[slug] = self.archives.detail(slug=slug)
        return self._archive_cache[slug].api_url

    def add_case_to_archive(
        self,
        *,
        archive_slug: str,
        values: list[SocketValueSpec],
    ) -> gcapi.models.ArchiveItemPost:
        """
        This function takes an archive slug and a list of socket value specs.
        It then creates a single archive item for the archive.

        Example:
            ```Python
            from gcapi import SocketValueSpec

            client.add_case_to_archive(
                archive_slug="i-am-an-archive",
                values=[
                    SocketValueSpec(socket_slug="report", file="report.pdf"),
                    SocketValueSpec(socket_slug="lung-volume", value=1.9),
                ],
            )
            ```

        ??? tip "Re-using existing images"
            Existing images on Grand Challenge can be re-used by either
            passing an API url, or an existing socket value:

            ```Python
            from gcapi import SocketValueSpec

            image = client.images.detail(pk="ad5...")
            ai = client.archive_items.detail(pk="f5...")
            socket_value = ai.values[0]

            values = [
                SocketValueSpec(socket_slug="slug-0", existing_image_api_url=image.api_url),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=socket_value),
                SocketValueSpec(socket_slug="slug-2", existing_image_api_url=socket_value.image),
            ]
            ```
        ??? tip "Re-using existing socket values"
            Existing socket values from other archive items can be re-used by
            passing a socket value. The sockets must be the same.

            For instance:

            ```Python
            from gcapi import SocketValueSpec

            ai = client.archive_items.detail(pk="f5...")
            values = [
                SocketValueSpec(socket_slug="slug-0", existing_socket_value=ai.values[0]),
                SocketValueSpec(socket_slug="slug-1", existing_socket_value=ai.values[1]),
                SocketValueSpec(socket_slug="slug-2", file="some_local_file"),
            ]
            ```

        Args:
            archive_slug: slug for the archive (e.g. `"i-am-an-archive"`).
                You can find this readily in the URL you use to visit the archive page:
                `https://grand-challenge.org/archives/i-am-an-archive/`

            values: A list of socket value specifications.
                Each specification defines a socket slug and exactly one source
                (`value`, `file`, `files`, `existing_image_api_url`, or
                `existing_socket_value`).

        Returns:
            The new archive item (post) object. Note that not all values will
                be immediately available until the background processing has completed.
        """
        archive_api_url = self._fetch_archive_api_url(archive_slug)

        try:
            created_archive_item = self._create_socket_value_set(
                creation_kwargs={"archive": archive_api_url},
                values=values,
                api=self.archive_items,
            )
        except SocketNotFound as e:
            raise ValueError(
                f"{e.slug} is not an existing socket. "
                f"Please provide one from this list: "
                f"https://grand-challenge.org/components/interfaces/inputs/"
            ) from e

        return cast(gcapi.models.ArchiveItemPost, created_archive_item)

    # Deprecated methods
    def add_cases_to_reader_study(self, *_, **__) -> Any:
        """
        !!! failure Deprecated
            Use `add_case_to_reader_study` instead. This method will be removed in a future version.
        """
        raise NotImplementedError(
            "add_cases_to_reader_study is no longer supported. Use the singular add_case_to_reader_study instead."
        )

    def add_cases_to_archive(self, *_, **__) -> Any:
        """
        !!! failure Deprecated
            Use `add_case_to_archive` instead. This method will be removed in a future version.
        """
        raise NotImplementedError(
            "add_cases_to_archive is no longer supported. Use the singular add_case_to_archive instead."
        )

    def run_external_job(self, *_, **__) -> Any:
        """
        !!! failure Deprecated
            Use `start_algorithm_job` instead. This method will be removed in a future version.
        """
        raise NotImplementedError(
            "run_external_job is no longer supported. Use start_algorithm_job instead."
        )

    def _fetch_socket_detail(
        self, slug: str
    ) -> gcapi.models.ComponentInterface:
        if slug not in self._socket_cache:
            try:
                self._socket_cache[slug] = self.interfaces.detail(slug=slug)
            except ObjectNotFound as e:
                raise SocketNotFound(slug=slug) from e
        return self._socket_cache[slug]

    def _create_socket_value_set(
        self,
        *,
        creation_kwargs: dict,
        values: list[SocketValueSpec],
        api: ModifiableMixin,
    ) -> SocketValuePostSet:

        with ExitStack() as stack:
            # Prepare the strategies
            strategies: list[SocketValueCreateStrategy] = []
            for spec in values:
                strategy = select_socket_value_strategy(
                    spec=spec,
                    client=self,
                )
                stack.enter_context(strategy)
                strategies.append(strategy)

            # Create the socket-value set
            socket_value_set = api.create(**creation_kwargs)

            # Update the socket-value set with the prepared values
            updated_socket_value_set = api.partial_update(
                pk=socket_value_set.pk,
                values=[s() for s in strategies],
            )

            return updated_socket_value_set

    def _update_socket_value_set(
        self,
        *,
        target_pk: str,
        values: list[SocketValueSpec],
        api: ModifiableMixin,
    ) -> SocketValuePostSet:

        with ExitStack() as stack:
            # Prepare the strategies
            strategies: list[SocketValueCreateStrategy] = []
            for spec in values:
                strategy = select_socket_value_strategy(
                    spec=spec,
                    client=self,
                )
                stack.enter_context(strategy)
                strategies.append(strategy)

            # Update the socket-value set with the prepared values
            return api.partial_update(
                pk=target_pk,
                values=[s() for s in strategies],
            )



from contextlib import nullcontext
from unittest.mock import MagicMock

import pytest
from grand_challenge_dicom_de_identifier.exceptions import (
    RejectedDICOMFileError,
)

from gcapi.create_strategies import (
    DICOMImageSetFileCreateStrategy,
    FileCreateStrategy,
    FileFromSVCreateStrategy,
    FileJSONCreateStrategy,
    ImageCreateStrategy,
    ImageFromImageCreateStrategy,
    ImageFromSVCreateStrategy,
    JobInputsCreateStrategy,
    SocketValueSpec,
    TooManyFiles,
    ValueCreateStrategy,
    ValueFromFileCreateStrategy,
    ValueFromSVStrategy,
    clean_file_source,
    select_socket_value_strategy,
)
from gcapi.exceptions import ObjectNotFound, SocketNotFound
from gcapi.models import AlgorithmInterface
from tests import TESTDATA
from tests.factories import (
    AlgorithmFactory,
    HyperlinkedComponentInterfaceValueFactory,
    HyperlinkedImageFactory,
    SocketFactory,
)


@pytest.mark.parametrize(
    "files,maximum_number,context",
    (
        (
            [TESTDATA / "test.json"],
            None,
            nullcontext(),
        ),
        (
            [str(TESTDATA / "test.json")],
            None,
            nullcontext(),
        ),
        (
            [TESTDATA / "test.json", TESTDATA / "test.json"],
            1,
            pytest.raises(TooManyFiles),
        ),
        (
            ["I DO NOT EXIST"],
            None,
            pytest.raises(FileNotFoundError),
        ),
        (
            object(),
            None,
            pytest.raises(
                TypeError, match="files must be a list of file paths"
            ),
        ),
        (
            [],
            None,
            pytest.raises(FileNotFoundError),
        ),
    ),
)
def test_prep_file_source(files, maximum_number, context):
    with context:
        clean_file_source(files, maximum_number=maximum_number)


file_socket = SocketFactory(super_kind="File")
image_socket = SocketFactory(super_kind="Image")
dicom_image_set_socket = SocketFactory(
    super_kind="Image",
    kind="DICOM Image Set",
)
value_socket = SocketFactory(super_kind="Value")


@pytest.mark.parametrize(
    "spec, socket, context, expected_cls",
    (
        (
            SocketValueSpec(
                socket_slug=file_socket.slug,
                files=[TESTDATA / "test.json"],
            ),
            file_socket,
            nullcontext(),
            FileCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=file_socket.slug,
                file=TESTDATA / "test.json",
            ),
            file_socket,
            nullcontext(),
            FileCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=file_socket.slug,
                file=[TESTDATA / "test.json"],  # type: ignore
            ),
            file_socket,
            pytest.raises(TypeError),
            None,
        ),
        (
            SocketValueSpec(
                socket_slug=file_socket.slug,
                files=[TESTDATA / "test.json", TESTDATA / "test.json"],
            ),
            file_socket,
            pytest.raises(
                ValueError,
                match="You can only provide one file",
            ),
            None,
        ),
        (
            SocketValueSpec(socket_slug=file_socket.slug, value=42),
            file_socket,
            nullcontext(),
            FileJSONCreateStrategy,
        ),
        (  # An existing socket value
            SocketValueSpec(
                socket_slug=file_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=file_socket,
                    file="https://grand-challenge.org/media/components/componentinterfacevalue/5d/57/1793367/file.json",
                ),
            ),
            file_socket,
            nullcontext(),
            FileFromSVCreateStrategy,
        ),
        (  # An existing socket value, but not with not the same socket
            SocketValueSpec(
                socket_slug=file_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=value_socket,
                    value=42,
                ),
            ),
            file_socket,
            pytest.raises(ValueError, match="does not match socket"),
            None,
        ),
        (  # An existing socket value, with a missing file (corrupt)
            SocketValueSpec(
                socket_slug=file_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=file_socket,
                    file=None,
                ),
            ),
            file_socket,
            pytest.raises(ValueError, match="must have a file"),
            None,
        ),
        (  # An invalid socket value
            SocketValueSpec(
                socket_slug=file_socket.slug,
                existing_socket_value="I am not a socket value",  # type: ignore
            ),
            file_socket,
            pytest.raises(ValueError, match="existing_socket_value must be a"),
            None,
        ),
    ),
)
def test_file_socket_value_strategy_init(spec, socket, context, expected_cls):
    client_mock = MagicMock()

    def mock_socket_detail(slug=None, **__):
        if slug == socket.slug:
            return socket
        else:
            return SocketFactory(slug=slug)

    client_mock._fetch_socket_detail = mock_socket_detail

    with context:
        with select_socket_value_strategy(
            spec=spec, client=client_mock
        ) as strategy:
            pass
    if expected_cls:
        assert type(strategy) is expected_cls


@pytest.mark.parametrize(
    "spec, socket, context, expected_cls",
    (
        (
            SocketValueSpec(
                socket_slug=image_socket.slug,
                files=[TESTDATA / "image10x10x101.mha"],
            ),
            image_socket,
            nullcontext(),
            ImageCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=image_socket.slug,
                file=TESTDATA / "image10x10x101.mha",
            ),
            image_socket,
            nullcontext(),
            ImageCreateStrategy,
        ),
        (  # DICOM Image Set, single file
            SocketValueSpec(
                socket_slug=dicom_image_set_socket.slug,
                file=TESTDATA / "basic.dcm",
                image_name="foo",
            ),
            dicom_image_set_socket,
            nullcontext(),
            DICOMImageSetFileCreateStrategy,
        ),
        (  # DICOM Image Set, multiple files
            SocketValueSpec(
                socket_slug=dicom_image_set_socket.slug,
                files=[TESTDATA / "basic.dcm"],
                image_name="foo",
            ),
            dicom_image_set_socket,
            nullcontext(),
            DICOMImageSetFileCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=image_socket.slug,
                files=[
                    TESTDATA / "image10x10x101.mha",
                    TESTDATA / "image10x10x101.mha",
                ],
            ),
            image_socket,
            nullcontext(),
            ImageCreateStrategy,
        ),
        (  # An existing socket value
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=image_socket,
                    image="https://example.test/api/v1/cases/images/a-uuid/",
                ),
            ),
            image_socket,
            nullcontext(),
            ImageFromSVCreateStrategy,
        ),
        (  # An existing socket value (DICOM Image Set)
            SocketValueSpec(
                socket_slug=dicom_image_set_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=dicom_image_set_socket,
                    image="https://example.test/api/v1/cases/images/a-uuid/",
                ),
            ),
            dicom_image_set_socket,
            nullcontext(),
            ImageFromSVCreateStrategy,
        ),
        (  # An existing socket value, but not with the same socket kind
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=value_socket,
                    value=42,
                ),
            ),
            image_socket,
            pytest.raises(ValueError, match="does not match socket"),
            None,
        ),
        (  # Same super kind but not exactly the same
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=SocketFactory(super_kind="Image"),
                    image="https://example.test/api/v1/cases/images/a-uuid/",
                ),
            ),
            image_socket,
            pytest.raises(ValueError, match="does not match socket"),
            None,
        ),
        (  # An existing socket value, with a missing image (corrupt)
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=image_socket,
                    image=None,
                ),
            ),
            image_socket,
            pytest.raises(ValueError, match="must have an image"),
            None,
        ),
        (  # An existing socket value, with an invalid value (corrupt)
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=None,  # type: ignore
            ),
            image_socket,
            pytest.raises(ValueError, match="existing_socket_value must be a"),
            None,
        ),
        (  # An existing image by API URL
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_image_api_url="https://example.test/api/v1/cases/images/a-uuid/",
            ),
            image_socket,
            nullcontext(),
            ImageFromImageCreateStrategy,
        ),
        (  # Non-existent or non-accessible image
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_image_api_url="I do not exist or you may not access me, search not",
            ),
            image_socket,
            pytest.raises(ObjectNotFound),
            None,
        ),
        (  # Non-existent or non-accessible image (via socket value)
            SocketValueSpec(
                socket_slug=image_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=image_socket,
                    image="I do not exist or you may not access me, search not",
                ),
            ),
            image_socket,
            pytest.raises(ObjectNotFound),
            None,
        ),
        (  # DICOM Image Set, missing image name
            SocketValueSpec(
                socket_slug=dicom_image_set_socket.slug,
                file=TESTDATA / "basic.dcm",
            ),
            dicom_image_set_socket,
            pytest.raises(
                ValueError, match="you must also specify an image_name"
            ),
            None,
        ),
        (  # DICOM Image Set, unsupported DICOM file
            SocketValueSpec(
                socket_slug=dicom_image_set_socket.slug,
                file=TESTDATA / "unsupported.dcm",
                image_name="foo",
            ),
            dicom_image_set_socket,
            pytest.raises(
                RejectedDICOMFileError,
                match="Unsupported SOP Class",
            ),
            None,
        ),
        (  # Image set, but providing image_name
            SocketValueSpec(
                socket_slug=image_socket.slug,
                files=[
                    TESTDATA / "image10x10x101.mha",
                    TESTDATA / "image10x10x101.mha",
                ],
                image_name="foo",
            ),
            image_socket,
            pytest.raises(
                ValueError,
                match="image_name can only be specified when uploading a DICOM image set",
            ),
            None,
        ),
    ),
)
def test_image_socket_value_strategy_init(spec, socket, context, expected_cls):
    client_mock = MagicMock()

    def mock_socket_detail(slug=None, **__):
        if slug == socket.slug:
            return socket
        else:
            raise SocketNotFound(slug=slug)

    client_mock._fetch_socket_detail = mock_socket_detail

    def mock_images_detail(api_url=None, **__):
        if api_url == "https://example.test/api/v1/cases/images/a-uuid/":
            return HyperlinkedImageFactory()
        else:
            raise ObjectNotFound  # Also covers non accessible images

    client_mock.images.detail = mock_images_detail

    with context:
        with select_socket_value_strategy(
            spec=spec, client=client_mock
        ) as strategy:
            pass
    if expected_cls:
        assert type(strategy) is expected_cls


@pytest.mark.parametrize(
    "spec, socket, context, expected_cls",
    (
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                files=[TESTDATA / "test.json"],
            ),
            value_socket,
            nullcontext(),
            ValueFromFileCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                file=TESTDATA / "test.json",
            ),
            value_socket,
            nullcontext(),
            ValueFromFileCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                files=[TESTDATA / "test.json", TESTDATA / "test.json"],
            ),
            value_socket,
            pytest.raises(ValueError, match="You can only provide one file"),
            None,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                value=42,
            ),
            value_socket,
            nullcontext(),
            ValueCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                value="String",
            ),
            value_socket,
            nullcontext(),
            ValueCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                value=["String"],
            ),
            value_socket,
            nullcontext(),
            ValueCreateStrategy,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                value=object(),  # Not JSON serializable
            ),
            value_socket,
            pytest.raises(ValueError, match="is not JSON serializable"),
            None,
        ),
        (
            SocketValueSpec(
                socket_slug=value_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=value_socket,
                    value=42,
                ),
            ),
            value_socket,
            nullcontext(),
            ValueFromSVStrategy,
        ),
        (  # Different socket
            SocketValueSpec(
                socket_slug=value_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=SocketFactory(super_kind="Value"),
                    value=42,
                ),
            ),
            value_socket,
            pytest.raises(ValueError, match="does not match socket"),
            None,
        ),
        (  # Corrupt - missing value
            SocketValueSpec(
                socket_slug=value_socket.slug,
                existing_socket_value=HyperlinkedComponentInterfaceValueFactory(
                    interface=value_socket,
                    value=None,
                ),
            ),
            value_socket,
            pytest.raises(ValueError, match="must have a value"),
            None,
        ),
        (  # Corrupt - missing value
            SocketValueSpec(
                socket_slug=value_socket.slug,
                existing_socket_value="Not a socket value",  # type: ignore
            ),
            value_socket,
            pytest.raises(ValueError, match="existing_socket_value must be a"),
            None,
        ),
    ),
)
def test_value_socket_value_strategy_init(spec, socket, context, expected_cls):
    client_mock = MagicMock()

    def mock_socket_detail(slug=None, **__):
        if slug == socket.slug:
            return socket
        else:
            return SocketFactory(slug=slug)

    client_mock._fetch_socket_detail = mock_socket_detail

    with context:
        with select_socket_value_strategy(
            spec=spec, client=client_mock
        ) as strategy:
            pass

    if expected_cls:
        assert type(strategy) is expected_cls


@pytest.mark.parametrize(
    "algorithm,inputs,context",
    (
        (  # algo socket < input socket
            AlgorithmFactory(interfaces=[]),
            [
                SocketValueSpec(
                    socket_slug="foo",
                    files=[TESTDATA / "image10x10x101.mha"],
                )
            ],
            pytest.raises(
                ValueError, match="No matching interface for sockets"
            ),
        ),
        (  # algo socket > input socket
            AlgorithmFactory(
                interfaces=[
                    AlgorithmInterface(
                        inputs=[
                            SocketFactory(),
                        ],
                        outputs=[],
                    )
                ]
            ),
            [],
            pytest.raises(
                ValueError, match="No matching interface for sockets"
            ),
        ),
        (
            # algo ci = input ci
            AlgorithmFactory(
                interfaces=[
                    AlgorithmInterface(
                        inputs=[
                            file_socket,
                        ],
                        outputs=[],
                    )
                ]
            ),
            [
                SocketValueSpec(
                    socket_slug=file_socket.slug,
                    files=[TESTDATA / "image10x10x101.mha"],
                )
            ],
            nullcontext(),
        ),
    ),
)
def test_job_inputs_create_prep(algorithm, inputs, context):
    client_mock = MagicMock()

    client_mock._fetch_socket_detail = MagicMock(return_value=file_socket)

    with context:
        with JobInputsCreateStrategy(
            algorithm=algorithm,
            inputs=inputs,
            client=client_mock,
        ):
            pass


@pytest.mark.parametrize(
    "spec_dict,context",
    (
        # Valid cases - exactly one value source
        (
            {"socket_slug": "test-socket", "value": 42},
            nullcontext(),
        ),
        (  # Value can be set to None
            {"socket_slug": "test-socket", "value": None},
            nullcontext(),
        ),
        (
            {"socket_slug": "test-socket", "files": [TESTDATA / "test.json"]},
            nullcontext(),
        ),
        (
            {
                "socket_slug": "test-socket",
                "files": [TESTDATA / "test.json"],
                "image_name": "a_image_name",
            },
            nullcontext(),
        ),
        (
            {
                "socket_slug": "test-socket",
                "existing_image_api_url": "https://example.com/api/v1/images/123/",
            },
            nullcontext(),
        ),
        (
            {
                "socket_slug": "test-socket",
                "existing_socket_value": HyperlinkedComponentInterfaceValueFactory(),
            },
            nullcontext(),
        ),
        # Invalid cases - multiple value sources
        (
            {"socket_slug": "test-socket", "value": 42, "files": []},
            pytest.raises(
                ValueError, match="Only one source can be specified"
            ),
        ),
        (
            {
                "socket_slug": "test-socket",
                "value": 42,
                "existing_image_api_url": "https://example.com/api/v1/images/123/",
            },
            pytest.raises(
                ValueError, match="Only one source can be specified"
            ),
        ),
        (
            {
                "socket_slug": "test-socket",
                "value": 42,
                "files": [],
                "existing_image_api_url": "https://example.test/api/v1/images/123/",
            },
            pytest.raises(
                ValueError, match="Only one source can be specified"
            ),
        ),
        (
            {
                "socket_slug": "test-socket",
                "files": [],
                "existing_image_api_url": "https://example.test/api/v1/images/123/",
                "existing_socket_value": HyperlinkedComponentInterfaceValueFactory(),
            },
            pytest.raises(
                ValueError, match="Only one source can be specified"
            ),
        ),
        # Invalid case - no value sources
        (
            {"socket_slug": "test-socket"},
            pytest.raises(
                ValueError, match="At least one source must be specified"
            ),
        ),
    ),
)
def test_socket_value_spec_validation(spec_dict, context):
    with context:
        SocketValueSpec(**spec_dict)


def test_dicom_image_set_file_create_strategy_closes_spools():
    spec = SocketValueSpec(
        socket_slug="dicom-image-set-socket",
        files=[
            TESTDATA / "basic.dcm",
            TESTDATA / "basic.dcm",
        ],
        image_name="foo",
    )
    socket = SocketFactory(
        slug="dicom-image-set-socket",
        super_kind="Image",
        kind="DICOM Image Set",
    )

    client_mock = MagicMock()

    def mock_socket_detail(slug=None, **__):
        if slug == socket.slug:
            return socket
        else:
            raise SocketNotFound(slug=slug)

    client_mock._fetch_socket_detail = mock_socket_detail

    with select_socket_value_strategy(
        spec=spec, client=client_mock
    ) as strategy:
        # A few sanity checks. Note: we do NOT call the strategy: exit it in
        # a half state
        assert isinstance(strategy, DICOMImageSetFileCreateStrategy), "Sanity"
        for f in strategy.content:
            assert not f.closed

    # After exiting the context, the spools should be closed
    for f in strategy.content:
        assert f.closed



import httpx
import pytest

from gcapi.retries import BaseRetryStrategy
from gcapi.transports import RetryTransport
from tests.utils import mock_transport_responses

MOCK_REQUEST = httpx.Request("GET", "https://example.test")
MOCK_RESPONSES = [
    httpx.Response(httpx.codes.NOT_FOUND),
    httpx.Response(httpx.codes.NOT_FOUND),
    httpx.Response(httpx.codes.OK),
]


class NoRetries(BaseRetryStrategy):
    @staticmethod
    def get_delay(*_, **__):
        return None


class EndlessRetries(BaseRetryStrategy):
    @staticmethod
    def get_delay(*_, **__):
        return 0


def test_invalid_retries():
    with pytest.raises(RuntimeError):
        RetryTransport(retry_strategy=object)  # type: ignore


def test_null_retries():
    transport = RetryTransport(retry_strategy=None)

    with mock_transport_responses(transport, MOCK_RESPONSES) as mock_info:
        response = transport.handle_request(request=MOCK_REQUEST)
        assert response is MOCK_RESPONSES[0]
        assert mock_info.num_requests == 1


def test_no_retry_strategy():
    transport = RetryTransport(retry_strategy=NoRetries)

    with mock_transport_responses(transport, MOCK_RESPONSES) as mock_info:
        response = transport.handle_request(request=MOCK_REQUEST)
        assert response is MOCK_RESPONSES[0]
        assert mock_info.num_requests == 1


def test_infinite_retry_strategy():
    transport = RetryTransport(retry_strategy=EndlessRetries)

    with mock_transport_responses(transport, MOCK_RESPONSES) as mock_info:
        response = transport.handle_request(request=MOCK_REQUEST)
        assert response is MOCK_RESPONSES[-1]
        assert mock_info.num_requests == len(MOCK_RESPONSES)



from __future__ import annotations

import json
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from tempfile import SpooledTemporaryFile
from typing import TYPE_CHECKING, Any

import httpx

if TYPE_CHECKING:
    from gcapi.client import Client

from grand_challenge_dicom_de_identifier.deidentifier import DicomDeidentifier

from gcapi.models import (
    Algorithm,
    ComponentInterface,
    ComponentInterfaceValuePostRequest,
    HyperlinkedComponentInterfaceValue,
    UserUpload,
)


class TooManyFiles(ValueError):
    pass


class UnsetType:
    pass


Unset = UnsetType()


@dataclass
class SocketValueSpec:
    """
    Specification for a socket value which provides exactly one source
    that will be used to create a socket value.

    Socket values are the items that constitute cases in job inputs, archive items,
    and display sets.

    Example:
        ```python
        from gcapi import SocketValueSpec

        # Directly from a value
        SocketValueSpec(socket_slug="my_value_socket", value=42)

        # Indirectly from a file
        SocketValueSpec(socket_slug="my_value_socket", file="/path/to/number.json")

        # Directly from a file
        SocketValueSpec(socket_slug="my_pdf_socket", file="/path/to/report.pdf")

        # Directly from multiple files
        SocketValueSpec(socket_slug="my_image_socket", files=["1.dcm", "2.dcm"])

        # Indirectly from an existing image
        SocketValueSpec(
            socket_slug="my_image_socket",
            existing_image_api_url="https://grand-challenge.org/api/v1/images/123/"
        )

        # Indirectly from an existing socket value
        display_set = client.display_sets.detail(pk="...")
        SocketValueSpec(
            socket_slug="my_value_socket",
            existing_socket_value=display_set.values[0]
        )

        # A DICOM Image set, including image_name
        display_set = client.display_sets.detail(pk="...")
        SocketValueSpec(
            socket_slug="my_value_socket",
            files=["/path/to/dicom/files/1.dcm", "/path/to/dicom/files/2.dcm"],
            image_name="My Image Name",
        )
        ```

    Args:
        socket_slug: The slug of the socket (required)

        value: Direct source value to set (can be None)
        file: Single source file path for file/image upload
            (e.g. `file="annotation.json"`)
        files: List of source file paths that constitute a single value/image
            (e.g. `files=["1.dcm", "2.dcm", "3.dcm"]`)
        image_name: Name for the image when uploading a DICOM image

        existing_image_api_url: An API URL of an existing image to reuse
        existing_socket_value: Existing socket value that can be reused
    """

    socket_slug: str

    value: Any = Unset
    file: str | Path | UnsetType = Unset
    files: list[str | Path] | UnsetType = Unset
    image_name: str | UnsetType = Unset

    existing_image_api_url: str | UnsetType = Unset
    existing_socket_value: HyperlinkedComponentInterfaceValue | UnsetType = (
        Unset
    )

    def __post_init__(self):

        # Wrap file
        if not isinstance(self.file, UnsetType):
            self.files = [self.file]
            del self.file

        self.validate()

    def validate(self) -> None:
        """Validate the spec without a socket context."""
        if not isinstance(self.socket_slug, str):
            raise TypeError("socket_slug must be a string")

        sources = []

        for field_name, field_value in self.__dict__.items():
            if field_name in ("socket_slug", "image_name"):
                continue
            if field_value is not Unset:
                sources.append(field_name)

        potential_source_fields = sorted(
            set(self.__dict__.keys()).difference({"socket_slug", "image_name"})
        )

        if len(sources) > 1:
            raise ValueError(
                f"Only one source can be specified, but got: {', '.join(sources)}. "
                f"Please specify only one of the available source fields ({', '.join(potential_source_fields)})."
            )

        if len(sources) == 0:
            raise ValueError(
                "At least one source must be specified. "
                "Please provide one of the available source fields ({', '.join(potential_source_fields)})."
            )

    def validate_against_socket(self, socket: ComponentInterface) -> None:
        """Validate the spec with the socket context."""
        if socket.slug != self.socket_slug:
            raise ValueError(
                f"Socket slug {self.socket_slug!r} does not match "
                f"the provided socket {socket.slug!r}"
            )

        self._validate_image_name_field(socket=socket)
        self._validated_existing_socket_value_field(socket=socket)

    def _validate_image_name_field(self, socket: ComponentInterface) -> None:
        if socket.kind.casefold() == "dicom image set":
            if all(
                [
                    self.image_name is Unset,
                    self.existing_socket_value is Unset,
                    self.existing_image_api_url is Unset,
                ]
            ):
                raise ValueError(
                    "When uploading a DICOM image set, "
                    "you must also specify an image_name."
                )
        elif self.image_name is not Unset:
            raise ValueError(
                "image_name can only be specified when "
                "uploading a DICOM image set."
            )

    def _validated_existing_socket_value_field(
        self,
        socket: ComponentInterface,
    ) -> None:
        if self.existing_socket_value is Unset:
            return

        socket_value = self.existing_socket_value
        if not isinstance(
            socket_value,
            HyperlinkedComponentInterfaceValue,
        ):
            raise ValueError(
                f"existing_socket_value must be a {HyperlinkedComponentInterfaceValue.__name__}"
            )

        if socket_value.interface.pk != socket.pk:
            raise ValueError(
                f"Source {socket_value.interface.title!r} does not "
                f"match socket {socket.slug!r}"
            )

    def __repr__(self) -> str:
        set_fields = []
        for field_name, field_value in self.__dict__.items():
            if field_name == "socket_slug":
                continue
            if field_value is not Unset:
                set_fields.append(field_name)

        parts = [f"socket_slug={self.socket_slug!r}"]
        for field_name in set_fields:
            field_value = getattr(self, field_name)
            parts.append(f"{field_name}={field_value!r}")
        return f"{self.__class__.__name__}({', '.join(parts)})"


def clean_file_source(
    files: list[Path | str], *, maximum_number: int | None = None
) -> list[Path]:
    if not isinstance(files, list):
        raise TypeError("files must be a list of file paths")

    # Ensure no more than can be handled
    if maximum_number is not None and len(files) > maximum_number:
        raise TooManyFiles(
            f"The maximum is {maximum_number}, "
            f"you provided {maximum_number}: {files}"
        )

    # Ensure items exist
    cleaned = []
    for p in [Path(f) for f in files]:
        if p.exists():
            cleaned.append(p)
        else:
            raise FileNotFoundError(p)
    if not cleaned:
        raise FileNotFoundError("No files provided")

    return cleaned


class BaseCreateStrategy:
    """
    Base class that describes strategies when creating items on Grand Challenge.

    Some items can directly be created via an api call. However, some need
    a chain of dependent items to exists. The raison d'être of the strategies
    is to re-use code that checks and handles these relations.

    Each strategy prepares when initiating to support batch preparing
    before things actually land on Grand Challenge. Preparing should **not**
    create anything on Grand Challenge. But can reach out to the API.

    Calling the strategy will, if applicable, upload contents and hence cause
    objects to be created on Grand Challenge.

    If for some reason, the strategy is aborted halfway, the `close` method
    can be called to clean up any resources held by the strategy.
    """

    client: Client

    def __init__(self, *, client):
        self.client = client

    def __call__(self) -> Any: ...

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        self.close()

    def close(self) -> None:
        """Close any resources held by the strategy."""
        pass


class SocketValueCreateStrategy(BaseCreateStrategy):
    def __init__(
        self,
        *,
        socket: ComponentInterface,
        client: Client,
    ) -> None:
        super().__init__(client=client)

        self.socket = socket

    def __call__(
        self: SocketValueCreateStrategy,
    ) -> ComponentInterfaceValuePostRequest:

        return ComponentInterfaceValuePostRequest(
            interface=self.socket.slug,
            value=None,
            file=None,
            image=None,
            upload_session=None,
            user_upload=None,
            user_uploads=None,
            image_name=None,
        )


def select_socket_value_strategy(  # noqa: C901
    *,
    spec: SocketValueSpec,
    client: Client,
) -> SocketValueCreateStrategy:

    socket = client._fetch_socket_detail(spec.socket_slug)
    kwargs = dict(
        socket=socket,
        client=client,
    )

    spec.validate_against_socket(socket=socket)

    if socket.super_kind.casefold() == "file":
        if not isinstance(spec.files, UnsetType):
            return FileCreateStrategy(
                files=spec.files,
                **kwargs,
            )
        elif not isinstance(spec.value, UnsetType):
            return FileJSONCreateStrategy(
                value=spec.value,
                **kwargs,
            )
        elif not isinstance(spec.existing_socket_value, UnsetType):
            return FileFromSVCreateStrategy(
                socket_value=spec.existing_socket_value,
                **kwargs,
            )
    elif socket.super_kind.casefold() == "image":
        if not isinstance(spec.files, UnsetType):
            if socket.kind.casefold() == "dicom image set" and not isinstance(
                spec.image_name, UnsetType
            ):
                return DICOMImageSetFileCreateStrategy(
                    files=spec.files,
                    image_name=spec.image_name,
                    **kwargs,
                )
            else:
                return ImageCreateStrategy(
                    files=spec.files,
                    **kwargs,
                )
        elif not isinstance(spec.existing_image_api_url, UnsetType):
            return ImageFromImageCreateStrategy(
                existing_image_api_url=spec.existing_image_api_url,
                **kwargs,
            )
        elif not isinstance(spec.existing_socket_value, UnsetType):
            return ImageFromSVCreateStrategy(
                existing_socket_value=spec.existing_socket_value,
                **kwargs,
            )
    elif socket.super_kind.casefold() == "value":
        if not isinstance(spec.files, UnsetType):
            return ValueFromFileCreateStrategy(
                files=spec.files,
                **kwargs,
            )
        elif not isinstance(spec.existing_socket_value, UnsetType):
            return ValueFromSVStrategy(
                existing_socket_value=spec.existing_socket_value,
                **kwargs,
            )
        elif not isinstance(spec.value, UnsetType):
            return ValueCreateStrategy(
                value=spec.value,
                **kwargs,
            )

    raise NotImplementedError(
        f"No strategy found that supports socket {socket.title} with spec {spec}"
    )


class FileCreateStrategy(SocketValueCreateStrategy):
    """Direct file upload strategy"""

    def __init__(
        self,
        *,
        files: list[str | Path],
        **kwargs,
    ):
        super().__init__(**kwargs)

        try:
            cleaned_files: list[Path] = clean_file_source(
                files,
                maximum_number=1,
            )
        except TooManyFiles as e:
            raise ValueError("You can only provide one file") from e

        self.content = cleaned_files[0]
        self.content_name = self.content.name

    def __call__(self):
        post_request = super().__call__()

        with open(self.content, "rb") as f:
            user_upload = self.client.uploads.upload_fileobj(
                fileobj=f, filename=self.content_name
            )

        post_request.user_upload = user_upload.api_url
        return post_request


class FileJSONCreateStrategy(SocketValueCreateStrategy):
    """Some JSON serializable Python object upload strategy"""

    def __init__(self, *, value: Any, **kwargs):
        super().__init__(**kwargs)

        try:
            json_bytes = json.dumps(value).encode()
        except TypeError:
            raise ValueError("Source is not JSON serializable") from None
        else:
            self.content = json_bytes
            self.content_name = self.socket.relative_path

    def __call__(self):
        post_request = super().__call__()

        with BytesIO(self.content) as f:
            user_upload = self.client.uploads.upload_fileobj(
                fileobj=f, filename=self.content_name
            )

        post_request.user_upload = user_upload.api_url
        return post_request


class FileFromSVCreateStrategy(SocketValueCreateStrategy):
    """Uses an existing SocketValue as source"""

    def __init__(
        self,
        *,
        socket_value: HyperlinkedComponentInterfaceValue,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if socket_value.file is None:
            raise ValueError(
                f"{HyperlinkedComponentInterfaceValue.__name__} must have a file"
            )

        self.content_api_url = socket_value.file
        self.content_name = Path(socket_value.file).name

    def __call__(self):
        post_request = super().__call__()

        # Cannot link a file directly to file, so we download it first

        content = self._download_from_socket()

        with BytesIO(content) as f:
            user_upload = self.client.uploads.upload_fileobj(
                fileobj=f, filename=self.content_name
            )

        post_request.user_upload = user_upload.api_url
        return post_request

    def _download_from_socket(self) -> bytes:
        response_or_json = self.client(
            url=self.content_api_url,
            follow_redirects=True,
        )
        if isinstance(response_or_json, httpx.Response):
            return response_or_json.content
        else:
            # Else, the response is actually the JSON already
            return json.dumps(response_or_json).encode()


class ImageCreateStrategy(SocketValueCreateStrategy):
    """Direct image-file upload strategy"""

    def __init__(
        self,
        *,
        files: list[str | Path],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.files = clean_file_source(files)

    def __call__(self):
        post_request = super().__call__()

        uploads: list[UserUpload] = []
        for file in self.files:
            with open(file, "rb") as f:
                uploads.append(
                    self.client.uploads.upload_fileobj(
                        fileobj=f, filename=file.name
                    )
                )

        post_request.user_uploads = [u.api_url for u in uploads]
        return post_request


class DICOMImageSetFileCreateStrategy(SocketValueCreateStrategy):

    _deidentifier_instance: DicomDeidentifier | None = None

    def __init__(
        self,
        *,
        files: list[str | Path],
        image_name: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.files = clean_file_source(files)
        self.image_name = image_name

        # Use the singleton deidentifier
        deidentifier = self.get_deidentifier()

        self.content = []
        try:
            for file in self.files:
                temp = SpooledTemporaryFile()
                # Add early to content so it gets closed on error
                self.content.append(temp)
                deidentifier.deidentify_file(file=file, output=temp)
                temp.seek(0)
        except Exception:
            self.close()
            raise

    def close(self):
        super().close()

        for temp in self.content:
            temp.close()

    @classmethod
    def get_deidentifier(cls) -> DicomDeidentifier:
        if cls._deidentifier_instance is None:
            cls._deidentifier_instance = DicomDeidentifier()
        return cls._deidentifier_instance

    def __call__(self):
        post_request = super().__call__()

        post_request.image_name = self.image_name

        uploads = []
        for idx, temp in enumerate(self.content):
            upload = self.client.uploads.upload_fileobj(
                fileobj=temp, filename=f"dicom_image_{idx}.dcm"
            )
            temp.close()
            uploads.append(upload.api_url)

        post_request.user_uploads = uploads

        return post_request


class ImageFromImageCreateStrategy(SocketValueCreateStrategy):
    """Indirect via an existing image"""

    def __init__(
        self,
        *,
        existing_image_api_url: str,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        api_url = str(existing_image_api_url)

        # assert it is a valid URL
        self.content_api_url = httpx.URL(api_url)

        # assert we have access to the image
        self.client.images.detail(api_url=api_url)

    def __call__(self):
        post_request = super().__call__()
        post_request.image = str(self.content_api_url)
        return post_request


class ImageFromSVCreateStrategy(SocketValueCreateStrategy):
    """Indirect via an existing SocketValue"""

    def __init__(
        self,
        *,
        existing_socket_value: HyperlinkedComponentInterfaceValue,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if existing_socket_value.image is None:
            raise ValueError(
                f"{HyperlinkedComponentInterfaceValue.__name__} must have an image"
            )

        api_url = existing_socket_value.image

        # assert it is a valid URL
        self.content_api_url = httpx.URL(api_url)

        # assert we have access to the image
        self.client.images.detail(api_url=api_url)

    def __call__(self):
        post_request = super().__call__()
        post_request.image = str(self.content_api_url)
        return post_request


class ValueFromFileCreateStrategy(SocketValueCreateStrategy):
    """Directly provided value"""

    def __init__(
        self,
        *,
        files: list[str | Path],
        **kwargs,
    ):
        super().__init__(**kwargs)

        try:
            cleaned_files = clean_file_source(files, maximum_number=1)
        except TooManyFiles as e:
            raise ValueError("You can only provide one file") from e

        self.content = cleaned_files[0]

        # Check parsable JSON, but for memory
        # considerations do not load it here yet
        with open(self.content) as fp:
            try:
                json.load(fp)
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"File {self.content} is not a valid JSON file. "
                    "Nothing has been uploaded."
                ) from e

    def __call__(self):
        post_request = super().__call__()

        with open(self.content) as fp:
            post_request.value = json.load(fp)

        return post_request


class ValueFromSVStrategy(SocketValueCreateStrategy):

    def __init__(
        self,
        *,
        existing_socket_value: HyperlinkedComponentInterfaceValue,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if existing_socket_value.value is None:
            raise ValueError(
                f"{HyperlinkedComponentInterfaceValue.__name__} must have a value"
            )

        self.content = existing_socket_value.value

    def __call__(self):
        post_request = super().__call__()
        post_request.value = self.content
        return post_request


class ValueCreateStrategy(SocketValueCreateStrategy):

    def __init__(self, *, value: Any, **kwargs):
        super().__init__(**kwargs)

        try:
            json.dumps(value)  # Check if it is JSON serializable
        except TypeError as e:
            raise ValueError(
                "Value is not JSON serializable. Nothing has been uploaded."
            ) from e

        self.content = value

    def __call__(self):
        post_request = super().__call__()
        post_request.value = self.content
        return post_request


class JobInputsCreateStrategy(BaseCreateStrategy):
    def __init__(
        self,
        *,
        algorithm: Algorithm,
        inputs: list[SocketValueSpec],
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.algorithm: Algorithm = algorithm
        self.input_strategies: list[SocketValueCreateStrategy] = []

        self._assert_matching_interface(inputs=inputs)

        try:
            for spec in inputs:
                socket_value_strategy = select_socket_value_strategy(
                    spec=spec,
                    client=self.client,
                )

                self.input_strategies.append(socket_value_strategy)
        except Exception:
            self.close()
            raise

    def close(self) -> None:
        for strategy in self.input_strategies:
            strategy.close()

    def _assert_matching_interface(
        self, *, inputs: list[SocketValueSpec]
    ) -> None:

        # Find a matching interface
        matching_interface = None

        input_socket_slugs = {spec.socket_slug for spec in inputs}

        best_matching_sockets_count = 0
        best_matching_interface = None
        for interface in self.algorithm.interfaces:
            interface_socket_slugs = {
                socket.slug for socket in interface.inputs
            }
            matching_count = len(input_socket_slugs & interface_socket_slugs)
            if matching_count == len(interface_socket_slugs):
                # All input sockets are present in the interface
                matching_interface = interface
                break
            else:
                if matching_count > best_matching_sockets_count:
                    best_matching_sockets_count = matching_count
                    best_matching_interface = {
                        socket.slug for socket in interface.inputs
                    }

        if matching_interface is None:
            msg = f"No matching interface for sockets {input_socket_slugs} could be found."
            if best_matching_interface is not None:
                msg += f" The closest match is {best_matching_interface}."
            raise ValueError(msg)

    def __call__(self):
        result = []
        for s in self.input_strategies:
            result.append(s())
            s.close()
        return result



class GCAPIError(Exception):
    """Base class for all exceptions"""


class ObjectNotFound(GCAPIError):  # noqa: N818
    """Zero objects found when one was expected"""


class SocketNotFound(ObjectNotFound):  # noqa: N818
    """A socket could not be found when one was expected"""

    def __init__(self, *args, slug, **kwargs):
        super().__init__(*args, **kwargs)
        self.slug = slug


class MultipleObjectsReturned(GCAPIError):  # noqa: N818
    """Multiple objects returned when one was expected"""



import logging
from collections.abc import Callable
from time import sleep

import httpx

from gcapi.retries import BaseRetryStrategy

logger = logging.getLogger(__name__)

Seconds = float


class BaseRetryTransport:
    def __init__(
        self,
        retry_strategy: Callable[..., BaseRetryStrategy] | None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.retry_strategy = retry_strategy
        if self.retry_strategy:
            obj = self.retry_strategy()
            if not isinstance(obj, BaseRetryStrategy):
                raise RuntimeError(
                    "Provided retries strategy should be None "
                    "or when called produce an instance of BaseRetry"
                )

    def _get_retry_delay(
        self,
        retry_strategy,
        response,
        request,
    ) -> tuple[BaseRetryStrategy | None, Seconds | None]:
        if self.retry_strategy is None:
            return None, None  # Noop
        elif retry_strategy is None:  # Initiate
            retry_strategy = self.retry_strategy()

        retry_delay = retry_strategy.get_delay(response)
        if retry_delay is not None:
            error_phrase = httpx.codes.get_reason_phrase(response.status_code)
            logger.error(
                f"{request.method} request to {request.url} failed with "
                f"{response.status_code} status ({error_phrase!r}): "
                f"retrying with {retry_delay}s delay"
            )

        return retry_strategy, retry_delay


class RetryTransport(BaseRetryTransport, httpx.HTTPTransport):
    """
    Transport that retries unsuccessful requests. Delays between retries are
    governed by a retry strategy.

    Once a request fails, a retries-strategy object is created from the provided
    retries class. This object is then queried to provide the delay in seconds
    via 'get_delay(response)'.

    The retry object should be an instance of BaseRetries.
    """

    def handle_request(
        self, request: httpx.Request, *args, **kwargs
    ) -> httpx.Response:
        retry_strategy = None
        while True:
            response = super().handle_request(request, *args, **kwargs)

            if response.is_success:
                break

            retry_strategy, retry_delay = self._get_retry_delay(
                retry_strategy,
                response,
                request,
            )

            if retry_delay is None:
                break
            else:
                # Close any connections kept open for this request
                response.close()
                sleep(retry_delay)

        return response



import warnings
from importlib.metadata import version as get_version

import httpx
from packaging import version


class UnsupportedVersionError(Exception):
    pass


def check_version(base_url):
    package_name = "gcapi"

    current_version = get_version(package_name)

    with httpx.Client() as client:
        response = client.get(f"{base_url}gcapi/")
        response.raise_for_status()

        api_data = response.json()

        latest_version = api_data["latest_version"]
        lowest_supported_version = api_data["lowest_supported_version"]

    current_version_v = version.parse(current_version)
    latest_version_v = version.parse(latest_version)
    lowest_supported_version_v = version.parse(lowest_supported_version)

    if current_version_v < lowest_supported_version_v:
        raise UnsupportedVersionError(
            f"You are using {package_name} version {current_version}. "
            f"However, the platform only supports {lowest_supported_version} "
            "or newer. Upgrade via `pip install --upgrade {package_name}`",
        )

    if current_version_v < latest_version_v:
        warnings.warn(
            f"You are using {package_name} version {current_version}. "
            f"However, version {latest_version} is available. You should consider"
            f" upgrading via `pip install --upgrade {package_name}`",
            UserWarning,
            stacklevel=0,
        )



import datetime
from unittest import mock

import pytest
from httpx import Response, codes

from gcapi.retries import SelectiveBackoffStrategy

NO_RETRIES = [None]


@pytest.mark.parametrize(
    "responses,delays",
    [
        ([Response(codes.OK)], NO_RETRIES),
        ([Response(codes.UNAUTHORIZED)], NO_RETRIES),
        (  # No retry server errors
            [
                Response(e)
                for e in (
                    codes.INTERNAL_SERVER_ERROR,
                    codes.NOT_IMPLEMENTED,
                    codes.HTTP_VERSION_NOT_SUPPORTED,
                    codes.VARIANT_ALSO_NEGOTIATES,
                    codes.LOOP_DETECTED,
                    codes.NOT_EXTENDED,
                    codes.NETWORK_AUTHENTICATION_REQUIRED,
                )
            ],
            NO_RETRIES * 7,
        ),
        *(  # Backoff responses
            (
                [Response(e)] * 10,
                [0.1, 0.2, 0.4, 0.8, 1.60, 3.20, 6.40, 12.8, None, None],
            )
            for e in (
                codes.BAD_GATEWAY,
                codes.SERVICE_UNAVAILABLE,
                codes.GATEWAY_TIMEOUT,
                codes.INSUFFICIENT_STORAGE,
            )
        ),
        (  # Mixed responses
            [
                Response(codes.BAD_GATEWAY),
                Response(codes.BAD_GATEWAY),
                Response(codes.GATEWAY_TIMEOUT),
                Response(codes.GATEWAY_TIMEOUT),
            ],
            [0.1, 0.2] * 2,
        ),
        # codes.TOO_MANY_REQUESTS without Retry-After header
        (
            [Response(codes.TOO_MANY_REQUESTS)] * 10,
            [0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, None, None],
        ),
        # codes.TOO_MANY_REQUESTS with Retry-After header with garbage
        (
            [
                Response(
                    codes.TOO_MANY_REQUESTS,
                    headers={"Retry-After": "foo"},
                )
            ]
            * 10,
            [0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, None, None],
        ),
        # codes.TOO_MANY_REQUESTS with Retry-After header as integer seconds
        (
            [
                Response(
                    codes.TOO_MANY_REQUESTS,
                    headers={"Retry-After": "120"},
                )
            ]
            * 10,
            [120.0] * 8 + [None, None],
        ),
        # codes.TOO_MANY_REQUESTS with Retry-After header as HTTP-date (in the past)
        (
            [
                Response(
                    codes.TOO_MANY_REQUESTS,
                    headers={"Retry-After": "Wed, 21 Oct 1990 07:28:00 GMT"},
                )
            ]
            * 2,
            [
                0,
                0,
            ],
        ),
        (
            # codes.TOO_MANY_REQUESTS with Retry-After header as HTTP-date
            [
                Response(
                    codes.TOO_MANY_REQUESTS,
                    headers={"Retry-After": "Fri, 01 Jan 1999 12:00:10 GMT"},
                )
            ]
            * 2,
            [
                10,
                10,
            ],
        ),
    ],
)
def test_selective_backoff_strategy(responses, delays):
    fixed_date = datetime.datetime(
        1999, 1, 1, 12, 0, 0, tzinfo=datetime.timezone.utc
    )
    with mock.patch("gcapi.retries.datetime") as mock_datetime:
        mock_datetime.datetime.now.return_value = fixed_date
        mock_datetime.datetime.side_effect = (
            lambda *args, **kwargs: datetime.datetime(*args, **kwargs)
        )
        generator = SelectiveBackoffStrategy(
            backoff_factor=0.1, maximum_number_of_retries=8
        )
        strategy = generator()
        for response, expected_delay in zip(responses, delays, strict=True):
            assert strategy.get_delay(response) == expected_delay


def test_example_retry_strategy_usage(docs_path):
    from examples.upload_retry_strategy import UploadRetryStrategy

    strategy = UploadRetryStrategy()

    delay = strategy.get_delay(
        Response(
            codes.BAD_REQUEST,
            json={
                "non_field_errors": [
                    "You have created too many uploads. Please try again later."
                ]
            },
        )
    )

    assert delay == 300



import datetime
from email.utils import parsedate_to_datetime

import httpx
from httpx import codes

Seconds = float


class BaseRetryStrategy:
    def get_delay(self, latest_response: httpx.Response) -> Seconds | None:
        """
        Returns the number of seconds to pause before the next retry,
        based on the latest response.

        Optionally, if None is returned no retry is to be performed.
        """
        raise NotImplementedError(
            "The 'get_delay' method must be implemented."
        )  # pragma: no cover


NO_RETRY = None


class SelectiveBackoffStrategy(BaseRetryStrategy):
    """
    Retries responses with codes of transient server errors (i.e. 5xx)
    with an exponential backoff.

    Each response code has its own backoff counter.
    """

    def __init__(self, backoff_factor: float, maximum_number_of_retries: int):
        self.backoff_factor = backoff_factor
        self.maximum_number_of_retries = maximum_number_of_retries
        self.earlier_number_of_retries: dict[int, int] = dict()

    def __call__(self) -> BaseRetryStrategy:
        return self.__class__(
            backoff_factor=self.backoff_factor,
            maximum_number_of_retries=self.maximum_number_of_retries,
        )

    def get_delay(self, latest_response: httpx.Response) -> Seconds | None:
        if latest_response.status_code in (
            codes.BAD_GATEWAY,
            codes.SERVICE_UNAVAILABLE,
            codes.GATEWAY_TIMEOUT,
            codes.INSUFFICIENT_STORAGE,
            codes.TOO_MANY_REQUESTS,
        ):
            return self._backoff_retries(latest_response)
        else:
            return NO_RETRY

    def _backoff_retries(self, response):
        num_retries = self.earlier_number_of_retries.get(
            response.status_code, 0
        )
        if num_retries >= self.maximum_number_of_retries:
            return NO_RETRY
        else:
            self.earlier_number_of_retries[response.status_code] = (
                num_retries + 1
            )
            delay = self.backoff_factor * (2**num_retries)

            if "Retry-After" in response.headers:
                retry_after = response.headers["Retry-After"]
                # The HTTP header comes in two formats: integer seconds and HTTP-date
                try:
                    # Try to parse as integer seconds
                    delay = float(retry_after)
                except ValueError:
                    # Try to parse as HTTP-date
                    try:
                        retry_after_dt = parsedate_to_datetime(retry_after)
                    except ValueError:
                        pass
                    else:
                        now = datetime.datetime.now(
                            datetime.timezone.utc
                        ).replace(tzinfo=retry_after_dt.tzinfo)
                        delay = (retry_after_dt - now).total_seconds()
                        delay = max(0, delay)

            return delay
